/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.List;
import java.util.Map;

import org.xmloperator.lambda3.exception.term.BrokenTermException;
import org.xmloperator.lambda3.exception.term.StillReferencedTermException;
import org.xmloperator.lambda3.impl.coding.Combinators;
import org.xmloperator.lambda3.impl.type.TermTypeFactory;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.TermRecycler;
import org.xmloperator.lambda3.model.Variable;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.type.TermType;
import org.xmloperator.lambda3.util.Stack;

/**
 * Abstraction implementation.
 */
final class AbstractionImpl extends TermImpl implements Abstraction {
	static int creationCount = 0;
	static int breakCount = 0;
	static final Stack<AbstractionImpl> reusableAbstractionStack = new Stack<AbstractionImpl>(
			AbstractionImpl.class, REUSE_CAPACITY);

	/**
	 * Constructs and returns a new Abstraction.
	 * 
	 * @param boundVariable
	 *            a free Variable.
	 * @param body
	 *            a Term body.
	 * @return
	 */
	static final AbstractionImpl newAbstraction(VariableImpl boundVariable,
			TermImpl body) {
		if (boundVariable == null || body == null
				|| boundVariable.abstraction != null)
			throw new IllegalArgumentException();
		AbstractionImpl abstraction = reusableAbstractionStack.pop();
		if (abstraction == null)
			abstraction = new AbstractionImpl();
		abstraction.boundVariable = boundVariable;
		abstraction.body = body;
		body.referenceCount++;
		boundVariable.setAbstraction(abstraction);
		creationCount++;
		return abstraction;
	}

	VariableImpl boundVariable;
	TermImpl body;
	String combinatorName = null;

	private AbstractionImpl() {
	}

	/**
	 * Replaces the body part of this Abstraction.
	 * 
	 * @param newBody
	 *            a new body.
	 */
	final void replaceBody(TermImpl newBody) {
		this.combinatorName = null;
		if (this.body != null)
			this.body.referenceCount--;
		this.body = newBody;
		if (this.body != null)
			this.body.referenceCount++;
	}

	/**
	 * Set the isClosed property as the same as the one of a given Abstraction.
	 * 
	 * @param abstraction
	 *            an Abstraction.
	 */
	final void setClosedLike(AbstractionImpl abstraction) {
		this.combinatorName = abstraction.combinatorName;
	}

	// ////////////////////
	// Term implementation
	// ////////////////////

	public final boolean isAbstraction() {
		return true;
	}

	public final boolean isBroken() {
		return this.body == null;
	}

	public final Term abstractionElimination(TermRecycler termRecycler) {
		return this.body.abstractionElimination(this.boundVariable,
				termRecycler);
	}

	// ///////////////////////////
	// Abstraction implementation
	// ///////////////////////////

	public final Variable getBoundVariable() {
		return this.boundVariable;
	}

	public final Term getBody() {
		return this.body;
	}

	public boolean isClosed() {
		return this.combinatorName != null;
	}

	public final String getCombinatorName() {
		return this.combinatorName != null && this.combinatorName.length() != 0 ? this.combinatorName
				: null;
	}

	public final Abstraction setCombinatorName(String name) {
		if (name == null)
			throw new IllegalArgumentException();
		if (this.combinatorName != null)
			throw new IllegalStateException();
		if (ASSERTION)
			AssertUtils.verifyClosure(this);
		this.combinatorName = name.intern();
		return this;
	}

	public final Term etaReduce() {
		if (!this.body.isApplication())
			return this;
		ApplicationImpl application = (ApplicationImpl) this.body;
		if (application.argument != this.boundVariable)
			return this;
		if (AssertUtils.referenceCountOf(application.function,
				this.boundVariable) != 0)
			return this;
		return application.function;
	}

	public final boolean isI() {
		return this.body == this.boundVariable;
	}

	public boolean isK() {
		if (!this.body.isAbstraction())
			return false;
		return ((AbstractionImpl) this.body).body == this.boundVariable;
	}

	public final boolean isF() {
		if (!this.body.isAbstraction())
			return false;
		AbstractionImpl abstraction = (AbstractionImpl) this.body;
		return abstraction.body == abstraction.boundVariable;
	}

	public final boolean isS() {
		if (!this.body.isAbstraction())
			return false;
		AbstractionImpl a2 = (AbstractionImpl) this.body;
		if (!a2.body.isAbstraction())
			return false;
		AbstractionImpl a3 = (AbstractionImpl) a2.body;
		if (!a3.body.isApplication())
			return false;
		ApplicationImpl x1x3_x2x3 = (ApplicationImpl) a3.body;
		if (!x1x3_x2x3.function.isApplication()
				|| !x1x3_x2x3.argument.isApplication())
			return false;
		ApplicationImpl x1x3 = (ApplicationImpl) x1x3_x2x3.function;
		ApplicationImpl x2x3 = (ApplicationImpl) x1x3_x2x3.argument;
		return x1x3.function == this.boundVariable
				&& x1x3.argument == a3.boundVariable
				&& x2x3.function == a2.boundVariable
				&& x2x3.argument == a3.boundVariable;
	}

	public boolean isB() {
		if (!this.body.isAbstraction())
			return false;
		AbstractionImpl a2 = (AbstractionImpl) this.body;
		if (!a2.body.isAbstraction())
			return false;
		AbstractionImpl a3 = (AbstractionImpl) a2.body;
		if (!a3.body.isApplication())
			return false;
		ApplicationImpl x1_x2x3 = (ApplicationImpl) a3.body;
		if (x1_x2x3.function != this.boundVariable
				|| !x1_x2x3.argument.isApplication())
			return false;
		ApplicationImpl x2x3 = (ApplicationImpl) x1_x2x3.argument;
		return x2x3.function == a2.boundVariable
				&& x2x3.argument == a3.boundVariable;
	}

	public final boolean isC() {
		if (!this.body.isAbstraction())
			return false;
		AbstractionImpl a2 = (AbstractionImpl) this.body;
		if (!a2.body.isAbstraction())
			return false;
		AbstractionImpl a3 = (AbstractionImpl) a2.body;
		if (!a3.body.isApplication())
			return false;
		ApplicationImpl x1x3_x2 = (ApplicationImpl) a3.body;
		if (!x1x3_x2.function.isApplication()
				|| x1x3_x2.argument != a2.boundVariable)
			return false;
		ApplicationImpl x1x3 = (ApplicationImpl) x1x3_x2.function;
		return x1x3.function == this.boundVariable
				&& x1x3.argument == a3.boundVariable;
	}

	public final boolean isY() {
		if (!this.body.isApplication())
			return false;
		ApplicationImpl application = (ApplicationImpl) this.body;
		return (application.function.isAbstraction()
				&& application.argument.isAbstraction()
				&& ((AbstractionImpl) application.function)
						.isYPart(this.boundVariable) && ((AbstractionImpl) application.argument)
				.isYPart(this.boundVariable));
	}

	private boolean isYPart(Variable x1) {
		if (!this.body.isApplication())
			return false;
		ApplicationImpl x1_x2x2 = (ApplicationImpl) this.body;
		if (x1_x2x2.function != x1 || !x1_x2x2.argument.isApplication())
			return false;
		ApplicationImpl x2x2 = (ApplicationImpl) x1_x2x2.argument;
		return x2x2.function == this.boundVariable
				&& x2x2.argument == this.boundVariable;
	}

	public final boolean isTuple() {
		TermImpl term = this.body;
		while (term.isApplication())
			term = ((ApplicationImpl) term).function;
		if (term != this.boundVariable)
			return false;
		term = this.body;
		while (term.isApplication()) {
			ApplicationImpl application = (ApplicationImpl) term;
			if (application.argument == this.boundVariable
					|| AssertUtils.referenceCountOf(application.argument,
							this.boundVariable) != 0)
				return false;
			term = application.function;
		}
		return true;
	}

	public final int getChurchNumeral() {
		if (!this.body.isAbstraction())
			return -1;
		AbstractionImpl abstraction = (AbstractionImpl) this.body;
		Term body = abstraction.body;
		int value = 0;
		while (body.isApplication()) {
			ApplicationImpl application = (ApplicationImpl) body;
			if (application.function != this.boundVariable)
				return -1;
			value++;
			body = application.argument;
		}
		return body == abstraction.boundVariable ? value : -1;
	}

	// //////////////////
	// TermImpl override
	// //////////////////

	final void breakIt() {
		if (this.referenceCount != 0)
			throw new StillReferencedTermException(this);
		if (this.body == null)
			throw new BrokenTermException(this);
		this.combinatorName = null;
		this.body.referenceCount--;
		this.body = null;
		this.boundVariable = null;
		reusableAbstractionStack.push(this);
		breakCount++;
	}

	final TermImpl abstractionElimination(Variable boundVariable,
			TermRecycler termRecycler) {
		if (AssertUtils.referenceCountOf(this, boundVariable) == 0)
			return ApplicationImpl.newApplication(
					(TermImpl) Combinators.newK(), this.body
							.abstractionElimination(this.boundVariable,
									termRecycler));
		if (this.body == boundVariable)
			/*
			 * T[\xy.x] => K
			 */
			return (TermImpl) Combinators.newK();
		if (this.body.isAbstraction()) {
			AbstractionImpl abstraction = (AbstractionImpl) this.body;
			if (abstraction.body.isApplication()) {
				ApplicationImpl application = (ApplicationImpl) abstraction.body;
				if (application.function == boundVariable
						&& application.argument.isApplication()) {
					ApplicationImpl applicationArgument = (ApplicationImpl) application.argument;
					if (applicationArgument.function == this.boundVariable
							&& applicationArgument.argument == abstraction.boundVariable)
						/*
						 * T[\xyz.x(yz)] => B
						 */
						return (TermImpl) Combinators.newB();
				} else if (application.function.isApplication()
						&& application.argument == this.boundVariable) {
					ApplicationImpl applicationFunction = (ApplicationImpl) application.function;
					if (applicationFunction.function == boundVariable
							&& applicationFunction.argument == abstraction.boundVariable)
						/*
						 * T[\xyz.xzy] => C
						 */
						return (TermImpl) Combinators.newC();
				} else if (application.function.isApplication()
						&& application.argument.isApplication()) {
					ApplicationImpl applicationFunction = (ApplicationImpl) application.function;
					ApplicationImpl applicationArgument = (ApplicationImpl) application.argument;
					if (applicationFunction.function == boundVariable
							&& applicationFunction.argument == abstraction.boundVariable
							&& applicationArgument.function == this.boundVariable
							&& applicationArgument.argument == abstraction.boundVariable)
						/*
						 * T[\xyz.xz(yz)] => S
						 */
						return (TermImpl) Combinators.newS();
				}
			}
		}
		TermImpl transformedBody = this.body.abstractionElimination(
				this.boundVariable, termRecycler);
		TermImpl result = transformedBody.abstractionElimination(boundVariable,
				termRecycler);
		termRecycler.breakTermDeeply(transformedBody);
		return result;
	}

	final TermType assignType(List<Variable> boundVariableStack,
			List<TermType> boundVariableTypeStack,
			Map<Term, TermType> fromSharedTermToType) {
		if (this.referenceCount > 1) {
			TermType abstractionType = fromSharedTermToType.get(this);
			if (abstractionType != null)
				return abstractionType;
		}
		TermType boundVariableType = TermTypeFactory.newVariableTermType();
		boundVariableStack.add(this.boundVariable);
		boundVariableTypeStack.add(boundVariableType);
		TermType bodyType = this.body.assignType(boundVariableStack,
				boundVariableTypeStack, fromSharedTermToType);
		{
			int index = boundVariableStack.size() - 1;
			boundVariableStack.remove(index);
			boundVariableTypeStack.remove(index);
		}
		TermType abstractionType = TermTypeFactory.newFunctionTermType(
				boundVariableType, bodyType);
		if (this.referenceCount > 1)
			fromSharedTermToType.put(this, abstractionType);
		return abstractionType;
	}
}
