/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.xmloperator.lambda3.exception.term.BrokenTermException;
import org.xmloperator.lambda3.exception.term.ExternalBoundVariableException;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Variable;

/**
 * Assertion utilities.
 */
public abstract class AssertUtils {

	/**
	 * Verify the closure of the given root Term. It must contain no Variable
	 * bound to an external Abstraction.
	 * 
	 * @throws ExternalBoundVariableException
	 *             if some Variable contained within the given root Term is
	 *             bound to an external Abstraction.
	 */
	public static final void verifyClosure(Term rootTerm) {
		if (rootTerm == null)
			throw new IllegalArgumentException();
		List<TermImpl> termStack = new ArrayList<TermImpl>();
		List<Integer> todoArgumentStack = new ArrayList<Integer>();
		List<Variable> boundVariableStack = new ArrayList<Variable>();
		TermImpl term = (TermImpl) rootTerm;
		int todoArgument = 0;
		boolean onReturn = false;
		while (true) {
			if (onReturn
					|| term.isVariable()
					|| (term.isConstant() && ((ConstantImpl<?>) term).argumentCount == 0)) {
				if (term.isVariable()) {
					if (!boundVariableStack.contains(term))
						throw new ExternalBoundVariableException(rootTerm,
								(Variable) term);
				} else if (term.isAbstraction())
					boundVariableStack.remove(boundVariableStack.size() - 1);
				if (termStack.isEmpty())
					return;
				int index = termStack.size() - 1;
				term = termStack.remove(index);
				todoArgument = todoArgumentStack.remove(index).intValue();
				onReturn = todoArgument == 0;
			} else if (term.isConstant()) {
				ConstantImpl<?> constant = (ConstantImpl<?>) term;
				if (todoArgument != constant.getArgumentCount() - 1) {
					termStack.add(term);
					todoArgumentStack.add(Integer.valueOf(todoArgument + 1));
				}
				term = (TermImpl) constant.getArgument(todoArgument);
				if (term == null)
					throw new BrokenTermException(rootTerm, constant);
				todoArgument = 0;
			} else if (term.isAbstraction()) {
				AbstractionImpl abstraction = (AbstractionImpl) term;
				boundVariableStack.add(abstraction.boundVariable);
				termStack.add(term);
				todoArgumentStack.add(Integer.valueOf(0));
				term = abstraction.body;
				if (term == null)
					throw new BrokenTermException(rootTerm, abstraction);
				todoArgument = 0;
			} else if (term.isApplication()) {
				ApplicationImpl application = (ApplicationImpl) term;
				if (todoArgument != 0) {
					term = application.argument;
					if (term == null)
						throw new BrokenTermException(rootTerm, application);
					todoArgument = 0;
				} else {
					termStack.add(term);
					todoArgumentStack.add(Integer.valueOf(1));
					term = application.function;
					if (term == null)
						throw new BrokenTermException(rootTerm, application);
					todoArgument = 0;
				}
			} else
				throw new IllegalStateException();
		}
	}

	/**
	 * Returns the reference count of a given referenced Term within a root
	 * Term, excluding it.
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param referencedTerm
	 *            a referenced Term.
	 * @return the reference count of a given referenced Term within a root
	 *         Term, excluding it.
	 */
	public static final int referenceCountOf(final Term rootTerm,
			final Term referencedTerm) {
		if (rootTerm == null || referencedTerm == null)
			throw new IllegalArgumentException();
		if (rootTerm == referencedTerm)
			return 0;
		Set<Term> explorationSet = new HashSet<Term>();
		List<TermImpl> termStack = new ArrayList<TermImpl>();
		TermImpl term = (TermImpl) rootTerm;
		int referenceCount = 0;
		while (true) {
			if (term.isVariable()
					|| (term.isConstant() && ((ConstantImpl<?>) term).argumentCount == 0)
					|| term == referencedTerm
					|| (term.referenceCount > 1 && explorationSet
							.contains(term))) {
				if (term == referencedTerm)
					referenceCount++;
				if (termStack.isEmpty())
					return referenceCount;
				int index = termStack.size() - 1;
				term = termStack.remove(index);
			} else if (term.isConstant()) {
				ConstantImpl<?> constant = (ConstantImpl<?>) term;
				for (int k = constant.getArgumentCount() - 1; k > 0; k--) {
					TermImpl argument = (TermImpl) constant.getArgument(k);
					if (argument == null)
						throw new BrokenTermException(rootTerm, constant);
					termStack.add(argument);
				}
				TermImpl argument = (TermImpl) constant.getArgument(0);
				if (argument == null)
					throw new BrokenTermException(rootTerm, constant);
				term = argument;
			} else if (term.isAbstraction()) {
				AbstractionImpl abstraction = (AbstractionImpl) term;
				if (abstraction.body == null)
					throw new BrokenTermException(rootTerm, abstraction);
				if (abstraction.referenceCount > 1)
					explorationSet.add(abstraction);
				term = abstraction.body;
			} else if (term.isApplication()) {
				ApplicationImpl application = (ApplicationImpl) term;
				if (application.function == null
						|| application.argument == null)
					throw new BrokenTermException(rootTerm, application);
				if (application.argument != application.function)
					termStack.add(application.argument);
				if (application.referenceCount > 1)
					explorationSet.add(application);
				term = application.function;
			} else
				throw new IllegalArgumentException();
		}
	}
}
