/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

import org.xmloperator.lambda3.exception.ParsingException;
import org.xmloperator.lambda3.impl.coding.Combinators;
import org.xmloperator.lambda3.impl.coding.Constants;
import org.xmloperator.lambda3.model.NumberCoding;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Parser;
import org.xmloperator.lambda3.model.TermPrinter;

/**
 * Basic lambda Term parser.
 */
public class ParserImpl implements Parser {
	final List<NumberCoding> numberCodingList = new ArrayList<NumberCoding>();

	/**
	 * Constructor.
	 */
	public ParserImpl() {
	}

	/**
	 * Recognizes a predefined lambda Term represented by a single char.
	 * 
	 * @param ch
	 *            a char that may represents a predefined lambda Term.
	 * @return a predefined lambda Term represented by a single char.
	 */
	public Term recognizePredefinedChar(char ch) {
		if (ch == TermPrinter.CHAR_I)
			return (TermImpl) Combinators.newI();
		if (ch == TermPrinter.CHAR_K)
			return (TermImpl) Combinators.newK();
		if (ch == TermPrinter.CHAR_F)
			return (TermImpl) Combinators.newF();
		if (ch == TermPrinter.CHAR_Y)
			return (TermImpl) Combinators.newY();
		if (ch == TermPrinter.CHAR_S)
			return (TermImpl) Combinators.newS();
		if (ch == TermPrinter.CHAR_B)
			return (TermImpl) Combinators.newB();
		if (ch == TermPrinter.CHAR_C)
			return (TermImpl) Combinators.newC();
		if (ch == TermPrinter.CHAR_J)
			return (TermImpl) Constants.constantJ;
		if (ch == TermPrinter.CHAR_P)
			return (TermImpl) Constants.constantP;
		if (ch == TermPrinter.CHAR_T)
			return (TermImpl) Constants.newConstantT();
		if (ch == TermPrinter.CHAR_Z)
			return (TermImpl) Constants.constantZ;
		if (ch == TermPrinter.CHAR_M)
			return (TermImpl) Constants.constantNorm;
		return null;
	}

	/**
	 * Parses a complete lambda Term.
	 * 
	 * @param in
	 *            a ParsingInput.
	 * @param identifierStack
	 *            a stack of bound Variable identifier.
	 * @param boundVariableStack
	 *            a stack of bound Variable.
	 * @return the parsed Term.
	 * @throws IOException
	 *             in case of input exception
	 */
	final TermImpl parse(ParsingInput in, List<String> identifierStack,
			List<VariableImpl> boundVariableStack) throws IOException {
		TermImpl term1 = this.parseUnaryTerm(in, identifierStack,
				boundVariableStack);
		if (term1 == null)
			in.throwsParsingException("Lambda term expected");
		while (true) {
			TermImpl term2 = this.parseUnaryTerm(in, identifierStack,
					boundVariableStack);
			if (term2 == null)
				return term1;
			term1 = ApplicationImpl.newApplication(term1, term2);
		}
	}

	/**
	 * Parses an unary Term, not an Application.
	 * 
	 * @param in
	 *            a ParsingInput.
	 * @param identifierStack
	 *            a stack of bound Variable identifier.
	 * @param boundVariableStack
	 *            a stack of bound Variable.
	 * @return the parsed Term.
	 * @throws IOException
	 *             in case of input exception
	 */
	final TermImpl parseUnaryTerm(ParsingInput in,
			List<String> identifierStack, List<VariableImpl> boundVariableStack)
			throws IOException {
		while (in.isChar(TermPrinter.CHAR_SPACE))
			in.next();
		char ch = in.getChar();
		if (ch == TermPrinter.CHAR_OPENED_PARENTHESIS) {
			in.next();
			TermImpl term = this.parse(in, identifierStack, boundVariableStack);
			in.parseChar(TermPrinter.CHAR_CLOSED_PARENTHESIS);
			return term;
		}
		if (in.isVariableIdentifier()) {
			String identifier = in.parseVariableIdentifier();
			int index = identifierStack.indexOf(identifier);
			if (index < 0)
				in.throwsParsingException("Free variable");
			VariableImpl boundVariable = boundVariableStack.get(index);
			return boundVariable;
		}
		if (ch == TermPrinter.CHAR_LAMBDA) {
			in.next();
			int boundVariableCount = 0;
			while (true) {
				boundVariableCount++;
				String identifier = in.parseVariableIdentifier();
				if (identifierStack.contains(identifier))
					in
							.throwsParsingException("Bound variable identifier already used");
				VariableImpl boundVariable = VariableImpl.newVariable();
				identifierStack.add(identifier);
				boundVariableStack.add(boundVariable);
				if (!in.isVariableIdentifier())
					break;
			}
			in.parseChar(TermPrinter.CHAR_POINT);
			TermImpl term = this.parse(in, identifierStack, boundVariableStack);
			for (int k = 0; k < boundVariableCount; k++) {
				int index = identifierStack.size() - 1;
				VariableImpl boundVariable = boundVariableStack.get(index);
				term = AbstractionImpl.newAbstraction(boundVariable, term);
				identifierStack.remove(index);
				boundVariableStack.remove(index);
			}
			return term;
		}
		if (ch == TermPrinter.CHAR_ONE_OF_N) {
			in.next();
			if (!in.isNumber())
				in.throwsParsingException("0..9 expected");
			int n = in.getChar() - '0';
			in.next();
			if (!in.isNumber())
				in.throwsParsingException("0..9 expected");
			int p = in.getChar() - '0';
			if (p > n)
				in.throwsParsingException("0.." + n + " expected");
			in.next();
			return (TermImpl) Combinators.newOneOfN(n, p);
		}
		if (ch == TermPrinter.CHAR_E) {
			in.next();
			if (!in.isNumber())
				in.throwsParsingException("1..3 expected");
			int n = in.getChar() - '0';
			if (n < 1 || n > 3)
				in.throwsParsingException("1..3 expected");
			in.next();
			return (TermImpl) Constants.newConstantTrilean(n);
		}
		for (NumberCoding numberCoding : this.numberCodingList)
			if (numberCoding.getIntroductionChar() == ch) {
				in.next();
				ch = in.getChar();
				boolean isSigned = ch == TermPrinter.CHAR_PLUS
						|| ch == TermPrinter.CHAR_MINUS;
				if (isSigned)
					in.next();
				TermImpl term;
				if (in.isNumber()) {
					int number = in.parseNumber();
					if (ch == TermPrinter.CHAR_MINUS)
						number = -number;
					term = (TermImpl) numberCoding.codeNumber(isSigned, number);
					if (term == null)
						in.throwsParsingException("Unimplemented number");
				} else {
					if (isSigned)
						in.throwsParsingException("Number expected");
					String functionName = in.parseExtensionIdentifier();
					term = (TermImpl) numberCoding.codeFunction(functionName);
					if (term == null)
						in.throwsParsingException("Unimplemented function");
				}
				if (in.isChar(TermPrinter.CHAR_SPACE))
					in.next();
				return term;
			}
		TermImpl predefinedTerm = (TermImpl) recognizePredefinedChar(ch);
		if (predefinedTerm != null)
			in.next();
		return predefinedTerm;
	}

	// ///////////////////////////
	// Parser implementation
	// ///////////////////////////

	public void addNumberCoding(NumberCoding numberCoding) {
		this.numberCodingList.add(numberCoding);
	}

	public final Term parse(Reader reader) throws IOException {
		ParsingInput in = new ParsingInput(reader);
		List<String> identifierStack = new ArrayList<String>();
		List<VariableImpl> boundVariableStack = new ArrayList<VariableImpl>();
		Term term = this.parse(in, identifierStack, boundVariableStack);
		if (!in.isEmpty())
			in.throwsParsingException("Lambda term expected");
		return term;
	}

	public final Term parse(String text) {
		try {
			return this.parse(new StringReader(text));
		} catch (IOException exception) {
			return null;
		} catch (ParsingException exception) {
			exception.setParsedText(text);
			throw exception;
		}
	}
}
