/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Function1;
import org.xmloperator.lambda3.model.Function2;
import org.xmloperator.lambda3.model.Function3;
import org.xmloperator.lambda3.model.Variable;
import org.xmloperator.lambda3.model.type.TermType;
import org.xmloperator.lambda3.model.type.TermTypeFormula;

/**
 * Lambda Term factory.
 * 
 * <pre>
 * For creating lambda terms:
 * 1.0: Uses TermFactory.newVariable() for creating free variables.
 * 1.1: Uses TermFactory.newConstant(Function1, TermType) for creating a one argument Constant.
 * 1.2: Uses TermFactory.newConstant(Function2, TermType, TermType) for creating a two arguments Constant.
 * 1.3: Uses TermFactory.newConstant(Function3, TermType, TermType, TermType) for creating a three arguments Constant.
 * 1.4: Uses TermFactory.newConstant(int) for creating an int Constant.
 * 2: Uses Term.applyTo() for creating applications.
 * 3: Uses Term.abstractOn() for creating abstractions.
 * </pre>
 */
public abstract class TermFactory {
	private static final ConstantInt[] constantInts = new ConstantInt[4096];

	/**
	 * Creates and returns a new Variable. It is free Variable, not bound to an
	 * Abstraction.
	 * 
	 * @return the created Variable.
	 */
	public static final Variable newVariable() {
		return VariableImpl.newVariable();
	}

	/**
	 * Creates and returns a new named Variable. It is free Variable, not bound
	 * to an Abstraction. Naming is used for debugging only.
	 * 
	 * @param nameChar
	 *            a char amongst 'a'..'w'.
	 * @return the created Variable.
	 */
	public static final Variable newVariable(char nameChar) {
		return VariableImpl.newVariable(nameChar);
	}

	/**
	 * Creates and returns a new Constant relative to a one argument function
	 * and a term type.
	 * 
	 * @return the created Constant.
	 */
	public static final Constant newConstant(Function1 function,
			TermType termType) {
		return Constant1Impl.newConstant(function, termType);
	}

	/**
	 * Creates and returns a new Constant relative to a one argument function
	 * and a term type formula.
	 * 
	 * @return the created Constant.
	 */
	public static final Constant newConstant(Function1 function,
			TermTypeFormula termTypeFormula) {
		return Constant1Impl.newConstant(function, termTypeFormula);
	}

	/**
	 * Creates and returns a new Constant relative to a two arguments function.
	 * 
	 * @return the created Constant.
	 */
	public static final Constant newConstant(Function2 function,
			TermType termType0, TermType termType1) {
		return Constant2Impl.newConstant(function, termType0, termType1);
	}

	/**
	 * Creates and returns a new Constant relative to a two arguments function.
	 * 
	 * @return the created Constant.
	 */
	public static final Constant newConstant(Function2 function,
			TermTypeFormula termTypeFormula0) {
		return Constant2Impl.newConstant(function, termTypeFormula0);
	}

	/**
	 * Creates and returns a new Constant relative to a three arguments
	 * function.
	 * 
	 * @return the created Constant.
	 */
	public static final Constant newConstant(Function3 function,
			TermType termType0, TermType termType1, TermType termType2) {
		return Constant3Impl.newConstant(function, termType0, termType1,
				termType2);
	}

	/**
	 * Creates and returns a new Constant relative to a three arguments
	 * function.
	 * 
	 * @return the created Constant.
	 */
	public static final Constant newConstant(Function3 function,
			TermTypeFormula termTypeFormula0) {
		return Constant3Impl.newConstant(function, termTypeFormula0);
	}

	/**
	 * Creates and returns a new ConstantInt.
	 * 
	 * @return the created ConstantInt.
	 */
	public static final ConstantInt newConstantInt(int intValue) {
		if (intValue < 0 || intValue >= constantInts.length)
			return new ConstantInt(intValue);
		ConstantInt constant = constantInts[intValue];
		if (constant == null) {
			constant = new ConstantInt(intValue);
			constantInts[intValue] = constant;
		}
		return constant;
	}

	/**
	 * Creates and returns a new ConstantFloat.
	 * 
	 * @return the created ConstantFloat.
	 */
	public static final ConstantFloat newConstantFloat(double floatValue) {
		return new ConstantFloat(floatValue);
	}
}
