/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.TermPrinter;
import org.xmloperator.lambda3.model.Variable;

/**
 * TermPrinter implementation.
 */
public class TermPrinterImpl implements TermPrinter {
	/**
	 * Returns a segment of underscore.
	 * 
	 * @param count
	 *            a non negative integer.
	 * @param underscoreChar
	 *            the underscore char.
	 * @return a segment of underscore.
	 */
	static final String underscore(int count, char underscoreChar) {
		if (count < 0)
			throw new IllegalArgumentException();
		StringBuffer sb = new StringBuffer();
		for (int k = 0; k < count; k++)
			sb.append(underscoreChar);
		return sb.toString();
	}

	/**
	 * Returns an underscore of an index.
	 * 
	 * @param index
	 *            a non negative integer.
	 * @param underscoreChar
	 *            the underscore char.
	 * @return an underscore of an index.
	 */
	static final String underscoreIndex(int index, char underscoreChar) {
		if (index < 0)
			throw new IllegalArgumentException();
		if (index <= 9)
			return String.valueOf(underscoreChar);
		int count = new Integer(index).toString().length();
		return underscore(count, underscoreChar);
	}

	final PrintWriter out;
	final Term underlinedTerm;
	final PrintOptions options = new PrintOptions();

	public TermPrinterImpl(PrintWriter out, Term underlinedTerm) {
		if (out == null)
			throw new IllegalArgumentException();
		this.out = out;
		this.underlinedTerm = underlinedTerm;
	}

	/**
	 * Prints a representation of a root Term if it is recognized as a shared
	 * Term or a predefined Term.
	 * <p>
	 * Example : "A1".
	 * </p>
	 * 
	 * @param rootTerm
	 *            a root TermImpl.
	 * @param boundVariableStack
	 *            a stack of bound Variable.
	 * @param sharedTermStack
	 *            a stack of shared Term.
	 * @param sharedTermContextStack
	 *            a stack of bound Variable context of shared Term.
	 * @param underscoreChar
	 *            if not null then prints an underscore.
	 * @return <tt>true</tt> if this Term has been recognized and printed,
	 *         <tt>false</tt> elsewhere.
	 */
	final boolean printRecognizedTerm(TermImpl rootTerm,
			List<Variable> boundVariableStack, List<TermImpl> sharedTermStack,
			List<List<Variable>> sharedTermContextStack, char underscoreChar) {
		{
			char predefinedIdentifier = this.predefinedIdentifier(rootTerm);
			if (predefinedIdentifier != 0) {
				this.out.print(underscoreChar != 0 ? underscoreChar
						: predefinedIdentifier);
				return true;
			}
		}
		if (rootTerm.referenceCount > 1) {
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_SHARED_TERM);
			int index = sharedTermStack.indexOf(rootTerm);
			if (index < 0) {
				index = sharedTermStack.size();
				sharedTermStack.add(rootTerm);
				sharedTermContextStack.add(new ArrayList<Variable>(
						boundVariableStack));
			}
			if (underscoreChar != 0)
				this.out.print(underscoreIndex(1 + index, underscoreChar));
			else
				this.out.print(1 + index);
			return true;
		}
		return false;
	}

	/**
	 * Prints a text representation of a Term as a tree using the De Bruijn
	 * indexed notation. If the underlinedTerm argument is not null then prints
	 * an underscore of the underlinedTerm within this Term.
	 * <p>
	 * Example : "\1".
	 * </p>
	 * 
	 * @param term
	 *            a Term.
	 * @param boundVariableStack
	 *            a stack of bound Variable.
	 * @param isUnderlined
	 *            if <tt>true</tt> then underlines this Term.
	 */
	final void printAsIndexed(Term term, List<Variable> boundVariableStack,
			boolean isUnderlined) {
		if (term.isVariable())
			this.printAsIndexed((VariableImpl) term, boundVariableStack,
					isUnderlined);
		else if (term.isConstant())
			this.printConstant((Constant) term, isUnderlined);
		else if (term.isAbstraction())
			this.printAsIndexed((AbstractionImpl) term, boundVariableStack,
					isUnderlined);
		else if (term.isApplication())
			this.printAsIndexed((ApplicationImpl) term, boundVariableStack,
					isUnderlined);
		else
			throw new IllegalStateException();
	}

	/**
	 * Prints a text representation of a Term as a tree. If the underlinedTerm
	 * argument is not null then prints an underscore of the underlinedTerm.
	 * <p>
	 * Example : "\x1.x1".
	 * </p>
	 * 
	 * @param term
	 *            a Term.
	 * @param boundVariableStack
	 *            a stack of bound Variable.
	 * @param isUnderlined
	 *            if <tt>true</tt> then underlines this Term.
	 */
	final void printAsLambda(Term term, List<Variable> boundVariableStack,
			boolean isUnderlined) {
		if (term.isVariable())
			this.printAsLambda((VariableImpl) term, boundVariableStack,
					isUnderlined);
		else if (term.isConstant())
			this.printConstant((Constant) term, isUnderlined);
		else if (term.isAbstraction())
			this.printAsLambda((AbstractionImpl) term, boundVariableStack,
					isUnderlined);
		else if (term.isApplication())
			this.printAsLambda((ApplicationImpl) term, boundVariableStack,
					isUnderlined);
		else
			throw new IllegalStateException();
	}

	/**
	 * Prints a text representation of a Term as a DAG. If the underlinedTerm
	 * argument is not null then prints an underscore of the underlinedTerm.
	 * <p>
	 * Example(Omega): "A1A1 A1=\x1.x1x1".
	 * </p>
	 * 
	 * @param term
	 *            a Term.
	 * @param boundVariableStack
	 *            a stack of bound Variable.
	 * @param sharedTermStack
	 *            a stack of shared Term.
	 * @param sharedTermContextStack
	 *            a stack of bound Variable context of shared Term.
	 * @param recognizeTerm
	 *            <tt>true</tt> for recognizing a shared or a predefined Term,
	 *            <tt>false</tt> for printing its content.
	 * @param isTuple
	 *            <tt>true</tt> for presenting this Term as a tuple part.
	 * @param isUnderlined
	 *            if <tt>true</tt> then underlines this Term.
	 */
	final void printAsDAG(Term term, List<Variable> boundVariableStack,
			List<TermImpl> sharedTermStack,
			List<List<Variable>> sharedTermContextStack, boolean recognizeTerm,
			boolean isTuple, boolean isUnderlined) {
		if (term.isVariable())
			this.printAsLambda((VariableImpl) term, boundVariableStack,
					isUnderlined);
		else if (term.isConstant())
			this.printAsDAG((Constant) term, boundVariableStack,
					sharedTermStack, sharedTermContextStack, recognizeTerm,
					isUnderlined);
		else if (term.isAbstraction())
			this.printAsDAG((AbstractionImpl) term, boundVariableStack,
					sharedTermStack, sharedTermContextStack, recognizeTerm,
					isTuple, isUnderlined);
		else if (term.isApplication())
			this.printAsDAG((ApplicationImpl) term, boundVariableStack,
					sharedTermStack, sharedTermContextStack, recognizeTerm,
					isTuple, isUnderlined);
		else
			throw new IllegalStateException();
	}

	// ///////////////
	// About Variable
	// ///////////////

	final void printAsIndexed(VariableImpl variable,
			List<Variable> boundVariableStack, boolean isUnderlined) {
		if (variable == this.underlinedTerm)
			isUnderlined = true;
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreChar = 0;
		if (variable.abstraction == null) {
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_UNNAMED_FREE_VARIABLE);
			return;
		}
		int index = boundVariableStack.indexOf(variable);
		if (index < 0) {
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_EXTERNAL_BOUND_VARIABLE);
			return;
		}
		index = boundVariableStack.size() - index;
		if (index > 9) {
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_UNDERSCORE);
			if (underscoreChar != 0)
				this.out.print(underscoreIndex(index, underscoreChar));
			else
				this.out.print(index);
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_SPACE);
		} else {
			if (underscoreChar != 0)
				this.out.print(underscoreChar);
			else
				this.out.print(index);
		}
	}

	final void printAsLambda(VariableImpl variable,
			List<Variable> boundVariableStack, boolean isUnderlined) {
		if (variable == this.underlinedTerm)
			isUnderlined = true;
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreChar = 0;
		{
			String name = variable.getName();
			if (name != null) {
				if (underscoreChar != 0)
					for (int k = 0; k < name.length(); k++)
						this.out.print(underscoreChar);
				else
					this.out.print(name);
				return;
			}
		}
		if (variable.abstraction == null) {
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_UNNAMED_FREE_VARIABLE);
			return;
		}
		int index = boundVariableStack.indexOf(variable);
		if (index < 0) {
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_EXTERNAL_BOUND_VARIABLE);
			return;
		}
		this.out.print(underscoreChar != 0 ? underscoreChar
				: TermPrinter.CHAR_BOUND_VARIABLE);
		if (underscoreChar != 0)
			this.out.print(underscoreIndex(1 + index, underscoreChar));
		else
			this.out.print(1 + index);
	}

	// ///////////////
	// About Constant
	// ///////////////

	final void printConstant(Constant constant, char underscoreChar) {
		String combinatorName = constant.getCombinatorName();
		String argsSummary = underscore(constant.getArgumentCount(),
				TermPrinter.CHAR_POINT);
		boolean useParenthesis = combinatorName.length() > 1
				|| argsSummary.length() > 0;
		if (useParenthesis)
			if (underscoreChar != 0)
				this.out.print(underscoreChar);
			else
				this.out.print(TermPrinter.CHAR_OPENED_PARENTHESIS);
		if (underscoreChar != 0)
			this.out.print(underscore(combinatorName.length()
					+ argsSummary.length(), underscoreChar));
		else {
			this.out.print(combinatorName);
			this.out.print(argsSummary);
		}
		if (useParenthesis)
			if (underscoreChar != 0)
				this.out.print(underscoreChar);
			else
				this.out.print(TermPrinter.CHAR_CLOSED_PARENTHESIS);
	}

	final void printConstant(Constant constant, boolean isUnderlined) {
		if (constant == this.underlinedTerm)
			isUnderlined = true;
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreChar = 0;
		this.printConstant(constant, underscoreChar);
	}

	final void printAsDAG(Constant constant, List<Variable> boundVariableStack,
			List<TermImpl> sharedTermStack,
			List<List<Variable>> sharedTermContextStack, boolean recognizeTerm,
			boolean isUnderlined) {
		if (constant == this.underlinedTerm)
			isUnderlined = true;
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreChar = 0;
		try {
			if (recognizeTerm
					&& this.printRecognizedTerm((TermImpl) constant,
							boundVariableStack, sharedTermStack,
							sharedTermContextStack, underscoreChar))
				return;
		} catch (Exception e) {
		}
		this.printConstant(constant, underscoreChar);
	}

	// //////////////////
	// About Abstraction
	// //////////////////

	final void printAsIndexed(AbstractionImpl abstraction,
			List<Variable> boundVariableStack, boolean isUnderlined) {
		if (abstraction == this.underlinedTerm)
			isUnderlined = true;
		boolean isBoundVariableUnderlined = isUnderlined
				|| (abstraction.boundVariable == this.underlinedTerm);
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		char underscoreBoundVariableChar = isBoundVariableUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreBoundVariableChar = 0;
		int abstractionCount = 0;
		while (true) {
			abstractionCount++;
			boundVariableStack.add(abstraction.boundVariable);
			this.out
					.print(underscoreBoundVariableChar != 0 ? underscoreBoundVariableChar
							: this.options.lambdaChar);
			if ((abstraction.body == null) || !abstraction.body.isAbstraction()
					|| !this.options.canAbstractionsBeGrouped)
				break;
			abstraction = (AbstractionImpl) abstraction.body;
		}
		if (abstraction.body == null)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_BROKEN_REFERENCE);
		else
			this.printAsIndexed(abstraction.body, boundVariableStack,
					isUnderlined);
		for (int k = 0; k < abstractionCount; k++)
			boundVariableStack.remove(boundVariableStack.size() - 1);
	}

	final void printAsLambda(AbstractionImpl abstraction,
			List<Variable> boundVariableStack, boolean isUnderlined) {
		if (abstraction == this.underlinedTerm)
			isUnderlined = true;
		boolean isBoundVariableUnderlined = isUnderlined
				|| (abstraction.boundVariable == this.underlinedTerm);
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		char underscoreBoundVariableChar = isBoundVariableUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null) {
			underscoreChar = 0;
			underscoreBoundVariableChar = 0;
		}
		this.out.print(underscoreChar != 0 ? underscoreChar
				: this.options.lambdaChar);
		int abstractionCount = 0;
		while (true) {
			abstractionCount++;
			boundVariableStack.add(abstraction.boundVariable);
			if (abstractionCount > 1
					&& this.options.separateIdentifiersWithSpaces)
				this.out.print(underscoreChar != 0 ? underscoreChar
						: TermPrinter.CHAR_SPACE);
			String name = abstraction.boundVariable.getName();
			if (name != null) {
				if (underscoreBoundVariableChar != 0)
					for (int k = 0; k < name.length(); k++)
						this.out.print(underscoreBoundVariableChar);
				else
					this.out.print(name);
			} else {
				this.out
						.print(underscoreBoundVariableChar != 0 ? underscoreBoundVariableChar
								: TermPrinter.CHAR_BOUND_VARIABLE);
				int index = boundVariableStack.size();
				if (underscoreBoundVariableChar != 0)
					this.out.print(underscoreIndex(index,
							underscoreBoundVariableChar));
				else
					this.out.print(index);
			}
			if ((abstraction.body == null) || !abstraction.body.isAbstraction()
					|| !this.options.canAbstractionsBeGrouped)
				break;
			abstraction = (AbstractionImpl) abstraction.body;
		}
		this.out.print(underscoreChar != 0 ? underscoreChar
				: TermPrinter.CHAR_POINT);
		if (abstraction.body == null)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_BROKEN_REFERENCE);
		else
			this.printAsLambda(abstraction.body, boundVariableStack,
					isUnderlined);
		for (int k = 0; k < abstractionCount; k++)
			boundVariableStack.remove(boundVariableStack.size() - 1);
	}

	final void printAsDAG(AbstractionImpl abstraction,
			List<Variable> boundVariableStack, List<TermImpl> sharedTermStack,
			List<List<Variable>> sharedTermContextStack, boolean recognizeTerm,
			boolean isTuple, boolean isUnderlined) {
		if (isTuple)
			throw new IllegalArgumentException();
		if (abstraction == this.underlinedTerm)
			isUnderlined = true;
		boolean isBoundVariableUnderlined = isUnderlined
				|| (abstraction.boundVariable == this.underlinedTerm);
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		char underscoreBoundVariableChar = isBoundVariableUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null) {
			underscoreChar = 0;
			underscoreBoundVariableChar = 0;
		}
		try {
			if (recognizeTerm
					&& this.printRecognizedTerm(abstraction,
							boundVariableStack, sharedTermStack,
							sharedTermContextStack, underscoreChar))
				return;
			{
				int churchNumeral = this.options.recognizeChurchNumerals ? abstraction
						.getChurchNumeral()
						: -1;
				if (churchNumeral >= 0) {
					this.out.print(underscoreChar != 0 ? underscoreChar
							: TermPrinter.CHAR_HASH);
					if (underscoreChar != 0)
						this.out.print(underscoreIndex(churchNumeral,
								underscoreChar));
					else
						this.out.print(churchNumeral);
					return;
				}
			}
			{
				String combinatorName = abstraction.getCombinatorName();
				if (combinatorName != null) {
					if (underscoreChar != 0)
						this.out.print(underscore(combinatorName.length() + 2,
								underscoreChar));
					else {
						this.out.print(TermPrinter.CHAR_OPENED_PARENTHESIS);
						this.out.print(combinatorName);
						this.out.print(TermPrinter.CHAR_CLOSED_PARENTHESIS);
					}
					return;
				}
			}
		} catch (Exception e) {
		}
		try {
			isTuple = this.options.canMakeTupples && abstraction.isTuple();
		} catch (Exception e) {
		}
		this.out.print(underscoreChar != 0 ? underscoreChar
				: isTuple ? TermPrinter.CHAR_LEFT_BRACKET
						: this.options.lambdaChar);
		int abstractionCount = 0;
		if (!isTuple) {
			while (true) {
				abstractionCount++;
				boundVariableStack.add(abstraction.boundVariable);
				if (abstractionCount > 1
						&& this.options.separateIdentifiersWithSpaces)
					this.out.print(underscoreChar != 0 ? underscoreChar
							: TermPrinter.CHAR_SPACE);
				String name = abstraction.boundVariable != null ? abstraction.boundVariable
						.getName()
						: null;
				if (name != null) {
					if (underscoreBoundVariableChar != 0)
						for (int k = 0; k < name.length(); k++)
							this.out.print(underscoreBoundVariableChar);
					else
						this.out.print(name);
				} else {
					this.out
							.print(underscoreBoundVariableChar != 0 ? underscoreBoundVariableChar
									: TermPrinter.CHAR_BOUND_VARIABLE);
					int index = boundVariableStack.size();
					if (underscoreBoundVariableChar != 0)
						this.out.print(underscoreIndex(index,
								underscoreBoundVariableChar));
					else
						this.out.print(index);
				}
				if ((abstraction.body == null)
						|| !abstraction.body.isAbstraction()
						|| !this.options.canAbstractionsBeGrouped
						|| this.predefinedIdentifier(abstraction.body) != 0
						|| (abstraction.body.isAbstraction() && ((AbstractionImpl) abstraction.body).combinatorName != null))
					break;
				abstraction = (AbstractionImpl) abstraction.body;
			}
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_POINT);
		}
		if (abstraction.body == null)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_BROKEN_REFERENCE);
		else
			this.printAsDAG(abstraction.body, boundVariableStack,
					sharedTermStack, sharedTermContextStack, true, isTuple,
					isUnderlined);
		for (int k = 0; k < abstractionCount; k++)
			boundVariableStack.remove(boundVariableStack.size() - 1);
		if (isTuple)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_RIGTH_BRACKET);
	}

	// //////////////////
	// About Application
	// //////////////////

	final void printAsIndexed(ApplicationImpl application,
			List<Variable> boundVariableStack, boolean isUnderlined) {
		if (application == this.underlinedTerm)
			isUnderlined = true;
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreChar = 0;
		boolean useParenthesisForFunction = (application.function != null)
				&& application.function.isAbstraction();
		boolean useParenthesisForArgument = (application.argument != null)
				&& !application.argument.isVariable();
		if (useParenthesisForFunction)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_OPENED_PARENTHESIS);
		if (application.function == null)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_BROKEN_REFERENCE);
		else
			this.printAsIndexed(application.function, boundVariableStack,
					isUnderlined);
		if (useParenthesisForFunction)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_CLOSED_PARENTHESIS);
		if (this.options.separateIdentifiersWithSpaces
				&& !useParenthesisForFunction && !useParenthesisForArgument)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_SPACE);
		if (useParenthesisForArgument)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_OPENED_PARENTHESIS);
		if (application.argument == null)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_BROKEN_REFERENCE);
		else
			this.printAsIndexed(application.argument, boundVariableStack,
					isUnderlined);
		if (useParenthesisForArgument)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_CLOSED_PARENTHESIS);
	}

	final void printAsLambda(ApplicationImpl application,
			List<Variable> boundVariableStack, boolean isUnderlined) {
		if (application == this.underlinedTerm)
			isUnderlined = true;
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreChar = 0;
		boolean useParenthesisForFunction = (application.function != null)
				&& application.function.isAbstraction();
		boolean useParenthesisForArgument = (application.argument != null)
				&& !application.argument.isVariable();
		if (useParenthesisForFunction)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_OPENED_PARENTHESIS);
		if (application.function == null)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_BROKEN_REFERENCE);
		else
			this.printAsLambda(application.function, boundVariableStack,
					isUnderlined);
		if (useParenthesisForFunction)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_CLOSED_PARENTHESIS);
		if (this.options.separateIdentifiersWithSpaces
				&& !useParenthesisForFunction && !useParenthesisForArgument)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_SPACE);
		if (useParenthesisForArgument)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_OPENED_PARENTHESIS);
		if (application.argument == null)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_BROKEN_REFERENCE);
		else
			this.printAsLambda(application.argument, boundVariableStack,
					isUnderlined);
		if (useParenthesisForArgument)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_CLOSED_PARENTHESIS);
	}

	final void printAsDAG(ApplicationImpl application,
			List<Variable> boundVariableStack, List<TermImpl> sharedTermStack,
			List<List<Variable>> sharedTermContextStack, boolean recognizeTerm,
			boolean isTuple, boolean isUnderlined) {
		if (application == this.underlinedTerm)
			isUnderlined = true;
		char underscoreChar = isUnderlined ? TermPrinter.CHAR_DASH
				: TermPrinter.CHAR_SPACE;
		if (this.underlinedTerm == null)
			underscoreChar = 0;
		try {
			if (recognizeTerm
					&& this.printRecognizedTerm(application,
							boundVariableStack, sharedTermStack,
							sharedTermContextStack, underscoreChar))
				return;
		} catch (Exception e) {
		}
		boolean useParenthesisForFunction = !isTuple
				&& application.function != null
				&& application.function.referenceCount == 1
				&& this.predefinedIdentifier(application.function) == 0
				&& application.function.isAbstraction()
				&& ((Abstraction) application.function).getCombinatorName() == null;
		boolean useParenthesisForArgument = !isTuple
				&& application.argument != null
				&& !application.argument.isVariable()
				&& application.argument.referenceCount == 1
				&& this.predefinedIdentifier(application.argument) == 0
				&& (!application.argument.isAbstraction() || ((Abstraction) application.argument)
						.getCombinatorName() == null);
		if (!isTuple || (application.function == null)
				|| !application.function.isVariable()) {
			if (useParenthesisForFunction)
				this.out.print(underscoreChar != 0 ? underscoreChar
						: TermPrinter.CHAR_OPENED_PARENTHESIS);
			if (application.function == null)
				this.out.print(underscoreChar != 0 ? underscoreChar
						: TermPrinter.CHAR_BROKEN_REFERENCE);
			else
				this.printAsDAG(application.function, boundVariableStack,
						sharedTermStack, sharedTermContextStack, true, isTuple,
						isUnderlined);
			if (useParenthesisForFunction)
				this.out.print(underscoreChar != 0 ? underscoreChar
						: TermPrinter.CHAR_CLOSED_PARENTHESIS);
		}
		if (isTuple && (application.function != null)
				&& !application.function.isVariable())
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_COMMA);
		if (!isTuple && this.options.separateIdentifiersWithSpaces
				&& !useParenthesisForFunction && !useParenthesisForArgument)
			this.out.print(underscoreChar != 0 ? underscoreChar
					: TermPrinter.CHAR_SPACE);
		{
			if (useParenthesisForArgument)
				this.out.print(underscoreChar != 0 ? underscoreChar
						: TermPrinter.CHAR_OPENED_PARENTHESIS);
			if (application.argument == null)
				this.out.print(underscoreChar != 0 ? underscoreChar
						: TermPrinter.CHAR_BROKEN_REFERENCE);
			else
				this.printAsDAG(application.argument, boundVariableStack,
						sharedTermStack, sharedTermContextStack, true, false,
						isUnderlined);
			if (useParenthesisForArgument)
				this.out.print(underscoreChar != 0 ? underscoreChar
						: TermPrinter.CHAR_CLOSED_PARENTHESIS);
		}
	}

	// ///////////////////////////
	// TermPrinter implementation
	// ///////////////////////////

	public void printAsIndexed(Term rootTerm) {
		this.printAsIndexed(rootTerm, new ArrayList<Variable>(), false);
	}

	public void printAsLambda(Term rootTerm) {
		this.printAsLambda(rootTerm, new ArrayList<Variable>(), false);
	}

	public char predefinedIdentifier(Term term) {
		if (term.isAbstraction()) {
			AbstractionImpl abstraction1 = (AbstractionImpl) term;
			if (!abstraction1.body.isAbstraction()) {
				if (abstraction1.body == abstraction1.boundVariable)
					return TermPrinter.CHAR_I;
				if (abstraction1.isY())
					return TermPrinter.CHAR_Y;
				return 0;
			}
			AbstractionImpl abstraction2 = (AbstractionImpl) abstraction1.body;
			if (!abstraction2.body.isAbstraction()) {
				if (abstraction1.isK())
					return TermPrinter.CHAR_K;
				if (abstraction1.isF())
					return TermPrinter.CHAR_F;
				return 0;
			}
			AbstractionImpl abstraction3 = (AbstractionImpl) abstraction2.body;
			if (!abstraction3.body.isAbstraction()) {
				if (abstraction1.isS())
					return TermPrinter.CHAR_S;
				if (abstraction1.isB())
					return TermPrinter.CHAR_B;
				if (abstraction1.isC())
					return TermPrinter.CHAR_C;
				return 0;
			}
			return 0;
		}
		if (term.isConstant()) {
			String name = ((Constant) term).getCombinatorName();
			if (name.length() > 1)
				return 0;
			char id = name.charAt(0);
			return (id >= 'A' && id <= 'Z') ? id : 0;
		}
		return 0;
	}

	public void printAsDAG(Term rootTerm) {
		List<TermImpl> sharedTermStack = new ArrayList<TermImpl>();
		List<List<Variable>> sharedTermContextStack = new ArrayList<List<Variable>>();
		this.printAsDAG(rootTerm, new ArrayList<Variable>(), sharedTermStack,
				sharedTermContextStack, true, false, false);
		for (int sharedTermIndex = 0; sharedTermIndex < sharedTermStack.size(); sharedTermIndex++) {
			TermImpl term = sharedTermStack.get(sharedTermIndex);
			List<Variable> boundVariableContext = sharedTermContextStack
					.get(sharedTermIndex);
			out.print(TermPrinter.CHAR_SEPARATOR);
			out.print(underlinedTerm != null ? TermPrinter.CHAR_SPACE
					: TermPrinter.CHAR_SHARED_TERM);
			if (underlinedTerm != null)
				out.print(underscoreIndex(1 + sharedTermIndex,
						TermPrinter.CHAR_SPACE));
			else
				out.print(1 + sharedTermIndex);
			out.print(underlinedTerm != null ? TermPrinter.CHAR_SPACE
					: TermPrinter.CHAR_EQUIVALENCE);
			this.printAsDAG(term, boundVariableContext, sharedTermStack,
					sharedTermContextStack, false, false, false);
		}
	}
}
