/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.xmloperator.lambda3.exception.term.NotClosedTermException;
import org.xmloperator.lambda3.exception.term.StillReferencedTermException;
import org.xmloperator.lambda3.impl.coding.Combinators;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.TermRecycler;
import org.xmloperator.lambda3.model.Variable;
import org.xmloperator.lambda3.model.type.TermType;
import org.xmloperator.lambda3.util.Stack;

/**
 * Variable implementation.
 */
@SuppressWarnings("unchecked")
final class VariableImpl extends TermImpl implements Variable {
	static int creationCount = 0;
	static int boundCount = 0;
	static int breakCount = 0;
	static final Stack<VariableImpl> reusableVariableStack = new Stack<VariableImpl>(
			VariableImpl.class, REUSE_CAPACITY);
	static final List<Variable>[] namedVariables = new List[23]; // 0..22 =
	// 'a'..'w'
	static final Map<Variable, Integer> fromVariableToVariableClass = new HashMap<Variable, Integer>();

	/**
	 * Constructs and returns a Variable, which is initially free.
	 * 
	 * @return the created Variable.
	 */
	static final VariableImpl newVariable() {
		creationCount++;
		VariableImpl variable = reusableVariableStack.pop();
		if (variable == null)
			return new VariableImpl();
		if (variable.referenceCount != 0)
			throw new StillReferencedTermException(variable);
		return variable;
	}

	/**
	 * Constructs and returns a named Variable, which is initially free.
	 * 
	 * @param nameChar
	 *            a char amongst 'a'..'w'.
	 * @return the created Variable.
	 */
	static final VariableImpl newVariable(char nameChar) {
		int index = nameChar - 'a';
		if (index < 0 || index >= namedVariables.length)
			throw new IllegalArgumentException();
		VariableImpl variable = newVariable();
		namedVariables[index].add(variable);
		fromVariableToVariableClass.put(variable, new Integer(index));
		return variable;
	}

	static {
		for (int k = 0; k < namedVariables.length; k++)
			namedVariables[k] = new ArrayList<Variable>();
	}

	Abstraction abstraction = null;

	private VariableImpl() {
	}

	/**
	 * Attach an Abstraction. This method is called par the Abstraction
	 * constructor.
	 * 
	 * @param abstraction
	 *            the Abstraction to attach.
	 */
	void setAbstraction(Abstraction abstraction) {
		if (abstraction == null)
			throw new IllegalArgumentException();
		if (this.abstraction != null)
			throw new IllegalStateException();
		this.abstraction = abstraction;
		boundCount++;
	}

	final String getName() {
		Integer variableClass = fromVariableToVariableClass.get(this);
		if (variableClass == null)
			return null;
		int classIndex = variableClass.intValue();
		char nameChar = (char) ('a' + classIndex);
		int instanceIndex = namedVariables[classIndex].indexOf(this);
		StringBuffer sb = new StringBuffer();
		sb.append(nameChar);
		sb.append(1 + instanceIndex);
		return sb.toString();
	}

	// ////////////////////
	// Term implementation
	// ////////////////////

	public final boolean isVariable() {
		return true;
	}

	public final boolean isBroken() {
		return false;
	}
	
	public final Term abstractionElimination(TermRecycler termRecycler) {
		return this;
	}

	// /////////////////////////////
	// Variable implementation
	// /////////////////////////////

	public final Abstraction getAbstraction() {
		return this.abstraction;
	}

	// //////////////////
	// TermImpl override
	// //////////////////

	final void breakIt() {
		if (this.referenceCount != 0)
			throw new StillReferencedTermException(this);
		this.abstraction = null;
		reusableVariableStack.push(this);
		breakCount++;
	}

	final TermImpl abstractionElimination(Variable boundVariable,
			TermRecycler termRecycler) {
		if (boundVariable == this)
			return (TermImpl) Combinators.newI();
		return ApplicationImpl.newApplication((TermImpl) Combinators.newK(),
				this);
	}

	final TermType assignType(List<Variable> boundVariableStack,
			List<TermType> boundVariableTypeStack,
			Map<Term, TermType> fromSharedTermToType) {
		int index = boundVariableStack.indexOf(this);
		if (index < 0)
			throw new NotClosedTermException(this);
		return boundVariableTypeStack.get(index);
	}
}
