/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.coding;

import org.xmloperator.lambda3.impl.TermFactory;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.NumberCoding;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Variable;

/**
 * Implementation of NumberCoding as Church numerals.
 * 
 * <pre>
 * Term("#n") = "\fx.(f^n)x"
 * Term("#+n") = "[K,#n]", including n=0
 * Term("#-n") = "[F,#n]", excepting n=0
 * Term("#-0") = "[K,#0]"
 * FunctionName ::= "isZero" | "succ" | "pred" | "add" | "mult" | "power"
 * 					| "SisZero" | "Ssucc" | "Spred" | "Smult"
 * </pre>
 */
public final class ChurchNumberCoding implements NumberCoding {
	public static final NumberCoding NUMBER_CODING = new ChurchNumberCoding();

	static final Variable newVariable() {
		return TermFactory.newVariable();
	}

	static final Variable newVariable(char nameChar) {
		return TermFactory.newVariable(nameChar);
	}

	/**
	 * Returns a new instance of Term(#0) = "\xy.y" = F.
	 * 
	 * @return a new instance of Term(#0) = "\xy.y" = F.
	 */
	public static final Abstraction newZero() {
		return Combinators.newF();
	}

	/**
	 * Returns a new instance of Term(#n) = "\fx.(f^n)x"
	 * 
	 * @param number
	 *            a positive or null number
	 * @return a new instance of Term(#n) = "\fx.(f^n)x"
	 */
	public static final Abstraction newNumber(int number) {
		if (number < 0)
			throw new IllegalArgumentException();
		Variable f = newVariable();
		Variable x = newVariable();
		Term body = x;
		for (int k = 0; k < number; k++)
			body = f.applyTo(body);
		return body.abstractOn(x).abstractOn(f).setCombinatorName("#" + number);
	}

	/**
	 * Returns a new instance of Term(#[+-]n)
	 * 
	 * @param number
	 *            a relative number
	 * @return a new instance of Term(#[+-]n)
	 */
	public static final Abstraction newSignedNumber(int number) {
		Term signTerm = Combinators.newBoolean(number >= 0);
		Term absoluteValueTerm = newNumber(number >= 0 ? number : -number);
		return Combinators.newPair(signTerm, absoluteValueTerm)
				.setCombinatorName("#" + (number >= 0 ? "+" : "") + number);
	}

	/**
	 * Returns a new instance of Term(#isZero) = "[\y.F,K]".
	 * 
	 * @return a new instance of Term(#isZero) = "[\y.F,K]".
	 */
	public static final Abstraction newIsZero() {
		Variable y = newVariable();
		return Combinators.newPair(Combinators.newF().abstractOn(y),
				Combinators.newK()).setCombinatorName("#isZero");
	}

	/**
	 * Returns a new instance of Term(#SisZero) = "\x.#isZero(xF)".
	 * 
	 * @return a new instance of Term(#SisZero) = "\x.#isZero(xF)".
	 */
	public static final Abstraction newSignedIsZero() {
		Variable x = newVariable();
		return newIsZero().applyTo(x.applyTo(Combinators.newF())).abstractOn(x)
				.setCombinatorName("#SisZero");
	}

	/**
	 * Returns a new instance of Term(#+0) = "[K,#0]".
	 * 
	 * @return a new instance of Term(#+0) = "[K,#0]".
	 */
	public static final Abstraction newSignedZero() {
		return Combinators.newPair(Combinators.newK(), newZero())
				.setCombinatorName("#+0");
	}

	/**
	 * Returns a new instance of the function "\x.#SisZero x[K,#0]x", which
	 * normalizes "#-0" in "#+0".
	 * 
	 * @return a new instance of the function "\x.#SisZero x[K,#0]x", which
	 *         normalizes "#-0" in "#+0".
	 */
	private static final Abstraction newNormalizeSign() {
		Variable x = newVariable();
		return newSignedIsZero().applyTo(x).applyTo(newSignedZero()).applyTo(x)
				.abstractOn(x).setCombinatorName("#normSign");
	}

	/**
	 * Returns a new instance of Term(#succ) = "\zxy.x(zxy)"
	 * 
	 * @return a new instance of Term(#succ) = "\zxy.x(zxy)"
	 */
	public static final Abstraction newSucc() {
		Variable x = newVariable();
		Variable y = newVariable();
		Variable z = newVariable();
		Term zxy = z.applyTo(x).applyTo(y);
		return x.applyTo(zxy).abstractOn(y).abstractOn(x).abstractOn(z)
				.setCombinatorName("#succ");
	}

	/**
	 * Returns a new instance of Term(#pred) = "\nfx.n(\gh.h(gf))(\m.x)(\m.m)"
	 * 
	 * @return a new instance of Term(#pred) = "\nfx.n(\gh.h(gf))(\m.x)(\m.m)"
	 */
	public static final Abstraction newPred() {
		Variable n = newVariable();
		Variable f = newVariable();
		Variable x = newVariable();
		Variable g = newVariable();
		Variable h = newVariable();
		Variable m = newVariable();
		Term t2 = h.applyTo(g.applyTo(f)).abstractOn(h).abstractOn(g);
		Term t3 = x.abstractOn(m);
		return (Abstraction) n.applyTo(t2).applyTo(t3).applyTo(
				Combinators.newI()).abstractOn(x).abstractOn(f).abstractOn(n)
				.setCombinatorName("#pred");
	}

	/**
	 * Returns a new instance of Term(#Ssucc)
	 * 
	 * <pre>
	 * "\x.xK[K,#succ (xF)](#Snormalize [F,#pred (xF)])"
	 * </pre>
	 * 
	 * @return a new instance of Term(#Ssucc)
	 */
	public static final Abstraction newSignedSucc() {
		Variable x = newVariable();
		Term caseK = Combinators.newPair(Combinators.newK(), newSucc().applyTo(
				x.applyTo(Combinators.newF())));
		Term caseF = newNormalizeSign().applyTo(
				Combinators.newPair(Combinators.newF(), newPred().applyTo(
						x.applyTo(Combinators.newF()))));
		return x.applyTo(Combinators.newK()).applyTo(caseK).applyTo(caseF)
				.abstractOn(x).setCombinatorName("#Ssucc");
	}

	/**
	 * Returns a new instance of Term(#Spred)
	 * 
	 * <pre>
	 * "\x.xK(#-isZero x[F,#1][K,#pred (xF)])[F,#succ (xF)]"
	 * </pre>
	 * 
	 * @return a new instance of Term(#Spred)
	 */
	public static final Abstraction newSignedPred() {
		Variable x = newVariable();
		Term caseK = newSignedIsZero().applyTo(x).applyTo(
				Combinators.newPair(Combinators.newF(), newNumber(1))).applyTo(
				Combinators.newPair(Combinators.newK(), newPred().applyTo(
						x.applyTo(Combinators.newF()))));
		Term caseF = Combinators.newPair(Combinators.newF(), newSucc().applyTo(
				x.applyTo(Combinators.newF())));
		return x.applyTo(Combinators.newK()).applyTo(caseK).applyTo(caseF)
				.abstractOn(x).setCombinatorName("#Spred");
	}

	/**
	 * Returns a new instance of Term(#add) = "\mnfx.mf(nfx)"
	 * 
	 * @return a new instance of Term(#add) = "\mnfx.mf(nfx)"
	 */
	public static final Abstraction newAdd() {
		Variable m = newVariable();
		Variable n = newVariable();
		Variable f = newVariable();
		Variable x = newVariable();
		Term nfx = n.applyTo(f).applyTo(x);
		return (Abstraction) m.applyTo(f).applyTo(nfx).abstractOn(x)
				.abstractOn(f).abstractOn(n).abstractOn(m).setCombinatorName(
						"#add");
	}

	/**
	 * Returns a new instance of Term(#mult) = "\mnf.n(mf)"
	 * 
	 * @return a new instance of Term(#mult) = "\mnf.n(mf)"
	 */
	public static final Abstraction newMult() {
		Variable m = newVariable();
		Variable n = newVariable();
		Variable f = newVariable();
		return n.applyTo(m.applyTo(f)).abstractOn(f).abstractOn(n)
				.abstractOn(m).setCombinatorName("#mult");
	}

	/**
	 * Returns a new instance of Term(#Smult)
	 * 
	 * <pre>
	 * "\xy.#Snormalize [xK(yK)(yKFK),#mult (xF)(yF)]"
	 * </pre>
	 * 
	 * @return a new instance of Term(#Smult)
	 */
	public static final Abstraction newSignedMult() {
		Variable x = newVariable();
		Variable y = newVariable();
		Term sign = x.applyTo(Combinators.newK()).applyTo(
				y.applyTo(Combinators.newK())).applyTo(
				y.applyTo(Combinators.newK()).applyTo(Combinators.newF())
						.applyTo(Combinators.newK()));
		Term product = newMult().applyTo(x.applyTo(Combinators.newF()))
				.applyTo(y.applyTo(Combinators.newF()));
		return newNormalizeSign().applyTo(Combinators.newPair(sign, product))
				.abstractOn(y).abstractOn(x).setCombinatorName("#Smult");
	}

	/**
	 * Returns a new instance of the succ operator of [#q, r] (see newDivByTwo).
	 * 
	 * @return a new instance of the succ operator of [#q, r] (see newDivByTwo).
	 */
	private static final Abstraction newDivByTwoSucc() {
		Abstraction K = Combinators.newK();
		Abstraction F = Combinators.newF();
		Variable p = newVariable();
		Term pK = p.applyTo(K);
		Abstraction caseK = Combinators.newPair(newSucc().applyTo(pK), F);
		Abstraction caseF = Combinators.newPair(pK, K);
		return p.applyTo(F).applyTo(caseK).applyTo(caseF).abstractOn(p)
				.setCombinatorName("#divByTwoSucc");
	}

	/**
	 * Returns a new instance of an operator #divByTwo that produces a pair
	 * [q,r] where q and r are respectively the quotient and the rest of the
	 * division by 2 of the Church numeral argument.
	 * 
	 * <pre>
	 * q is a Church numeral.
	 * r is a boolean (F=false=0=even, K=true=1=odd).
	 * </pre>
	 * 
	 * @return a new instance of an operator #divByTwo that produces a pair
	 *         [q,r] where q and r are respectively the quotient and the rest of
	 *         the division by 2 of the Church numeral argument.
	 */
	public static final Abstraction newDivByTwo() {
		Variable n = newVariable();
		Abstraction pZero = Combinators.newPair(newZero(), Combinators.newF());
		return n.applyTo(newDivByTwoSucc()).applyTo(pZero).abstractOn(n)
				.setCombinatorName("#divByTwo");
	}

	/**
	 * Returns a new instance of Term(#power) = "\mn.nm"
	 * 
	 * @return a new instance of Term(#power) = "\mn.nm"
	 */
	public static final Abstraction newPower() {
		Variable m = newVariable();
		Variable n = newVariable();
		return n.applyTo(m).abstractOn(n).abstractOn(m).setCombinatorName(
				"#power");
	}

	/**
	 * Returns a new instance of the Syracuse conjecture operator.
	 * 
	 * <pre>
	 * Input : [q,r] of divByTwo
	 * Output : number
	 * [n,F] -> n
	 * [n,K] -> 3*n+2
	 * </pre>
	 * 
	 * @return a new instance of the Syracuse conjecture operator.
	 */
	public static final Abstraction newSyracuse() {
		Variable p = newVariable();
		Term pK = p.applyTo(Combinators.newK());
		Term pF = p.applyTo(Combinators.newF());
		Term caseK = newAdd().applyTo(newNumber(2)).applyTo(
				newMult().applyTo(newNumber(3)).applyTo(pK));
		return pF.applyTo(caseK).applyTo(pK).abstractOn(p).setCombinatorName(
				"#syracuse");
	}

	/**
	 * Returns a new instance of the operator that gives the step count of the
	 * Syracuse sequence starting at a given number.
	 * 
	 * <pre>
	 * stepCount := 0
	 * loop
	 * 		[q,r] := divByTwo(n)
	 * 		if q==0 then return(stepCount)
	 * 		n := syracuse([q,r])
	 * 		stepCount += 1
	 * endloop
	 * </pre>
	 * 
	 * @return a new instance of the operator that gives the step count of the
	 *         Syracuse sequence starting at a given number.
	 */
	public static final Term newSyracuseStepCount() {
		Variable c = newVariable();
		Variable p = newVariable();
		Term pK = p.applyTo(Combinators.newK());
		Abstraction countP = newIsZero()
				.applyTo(pK)
				.applyTo(newZero())
				.applyTo(newSucc().applyTo(c.applyTo(newSyracuse().applyTo(p))))
				.abstractOn(p);
		Variable n = newVariable();
		Abstraction countN = countP.applyTo(newDivByTwo().applyTo(n))
				.abstractOn(n);
		Term G = countN.abstractOn(c).setCombinatorName("#syracuseStepCount");
		return Combinators.newY().applyTo(G);
	}

	// ////////////////////////////
	// NumberCoding implementation
	// ////////////////////////////

	public char getIntroductionChar() {
		return NumberCoding.CHURCH;
	}

	public Term codeNumber(boolean isSigned, int number) {
		return isSigned ? newSignedNumber(number) : newNumber(number);
	}

	public Term codeFunction(String functionName) {
		if (functionName.equals(ISZERO))
			return newIsZero();
		if (functionName.equals(SUCC))
			return newSucc();
		if (functionName.equals(PRED))
			return newPred();
		if (functionName.equals(ADD))
			return newAdd();
		if (functionName.equals(MULT))
			return newMult();
		if (functionName.equals(DIV_BY_TWO))
			return newDivByTwo();
		if (functionName.equals(POWER))
			return newPower();
		if (functionName.equals(SIGNED_ISZERO))
			return newSignedIsZero();
		if (functionName.equals(SIGNED_SUCC))
			return newSignedSucc();
		if (functionName.equals(SIGNED_PRED))
			return newSignedPred();
		if (functionName.equals(SIGNED_MULT))
			return newSignedMult();
		if (functionName.equals("syracuse"))
			return newSyracuse();
		if (functionName.equals("syracuseStepCount"))
			return newSyracuseStepCount();
		return null;
	}
}
