/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.coding;

import org.xmloperator.lambda3.impl.TermFactory;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Variable;

/**
 * Usual combinators.
 */
public abstract class Combinators {

	private static final Variable newVariable() {
		return TermFactory.newVariable();
	}

	/**
	 * Returns a new instance of the Term I = \x.x
	 * 
	 * @return a new instance of the Term I = \x.x
	 */
	public static final Abstraction newI() {
		Variable x = newVariable();
		return x.abstractOn(x).setCombinatorName("I");
	}

	/**
	 * Returns a new instance of the Term K = \xy.x
	 * 
	 * @return a new instance of the Term K = \xy.x
	 */
	public static final Abstraction newK() {
		Variable x = newVariable();
		Variable y = newVariable();
		return x.abstractOn(y).abstractOn(x).setCombinatorName("K");
	}

	/**
	 * Returns a new instance of the Term F = \xy.y
	 * 
	 * @return a new instance of the Term F = \xy.y
	 */
	public static final Abstraction newF() {
		Variable x = newVariable();
		Variable y = newVariable();
		return y.abstractOn(y).abstractOn(x).setCombinatorName("F");
	}

	/**
	 * Returns a new instance of a boolean Term.
	 * 
	 * @param b
	 *            a boolean.
	 * @return a new instance of a boolean Term.
	 */
	public static final Abstraction newBoolean(boolean b) {
		return b ? newK() : newF();
	}

	/**
	 * Returns a new instance of the Term S = \xyz.xz(yz)
	 * 
	 * @return a new instance of the Term S = \xyz.xz(yz)
	 */
	public static final Abstraction newS() {
		Variable x = newVariable();
		Variable y = newVariable();
		Variable z = newVariable();
		Term xz = x.applyTo(z);
		Term yz = y.applyTo(z);
		return xz.applyTo(yz).abstractOn(z).abstractOn(y).abstractOn(x)
				.setCombinatorName("S");
	}

	/**
	 * Returns a new instance of the Term B = \xyz.x(yz)
	 * 
	 * @return a new instance of the Term B = \xyz.x(yz)
	 */
	public static final Abstraction newB() {
		Variable x = newVariable();
		Variable y = newVariable();
		Variable z = newVariable();
		Term yz = y.applyTo(z);
		return x.applyTo(yz).abstractOn(z).abstractOn(y).abstractOn(x)
				.setCombinatorName("B");
	}

	/**
	 * Returns a new instance of the Term C = \xyz.xzy
	 * 
	 * @return a new instance of the Term C = \xyz.xzy
	 */
	public static final Abstraction newC() {
		Variable x = newVariable();
		Variable y = newVariable();
		Variable z = newVariable();
		Term xz = x.applyTo(z);
		return xz.applyTo(y).abstractOn(z).abstractOn(y).abstractOn(x)
				.setCombinatorName("C");
	}

	/**
	 * Returns a new instance of the Term Y = \x.(\y.x(yy))(\y.x(yy))
	 * 
	 * @return a new instance of the Term Y = \x.(\y.x(yy))(\y.x(yy))
	 */
	public static final Abstraction newY() {
		Variable x = newVariable();
		Variable y = newVariable();
		Abstraction z = x.applyTo(y.applyTo(y)).abstractOn(y);
		return z.applyTo(z).abstractOn(x).setCombinatorName("Y");
	}

	/**
	 * Returns a new Term "\x0x1..xn.xp"
	 * 
	 * @param n
	 *            a positive or null integer.
	 * @param p
	 *            a positive or null integer, lesser or equal to n.
	 * @return a new Term "\x0x1..xn.xp"
	 */
	public static final Abstraction newOneOfN(int n, int p) {
		if (n < 0 || p < 0 || p > n)
			throw new IllegalArgumentException();
		Variable xp = newVariable();
		Term term = xp;
		for (int k = n; k > p; k--)
			term = term.abstractOn(newVariable());
		term = term.abstractOn(xp);
		for (int k = p - 1; k >= 0; k--)
			term = term.abstractOn(newVariable());
		return ((Abstraction) term).setCombinatorName((p + 1) + "/" + (n + 1));
	}

	/**
	 * Returns a new instance of the pair [A,B] = "\x.xAB".
	 * 
	 * @param term1
	 *            a first Term.
	 * @param term2
	 *            a second Term.
	 * @return a pair of the two given terms.
	 */
	public static final Abstraction newPair(Term term1, Term term2) {
		Variable x = newVariable();
		return x.applyTo(term1).applyTo(term2).abstractOn(x);
	}

	/**
	 * Returns a new instance of the triplet [A,B,C] = "\x.xABC".
	 * 
	 * @param term1
	 *            a first Term.
	 * @param term2
	 *            a second Term.
	 * @param term3
	 *            a third Term.
	 * @return a triplet of the three given terms.
	 */
	public static final Abstraction newTriplet(Term term1, Term term2,
			Term term3) {
		Variable x = newVariable();
		return x.applyTo(term1).applyTo(term2).applyTo(term3).abstractOn(x);
	}
}
