/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.coding;

import org.xmloperator.lambda3.impl.TermFactory;
import org.xmloperator.lambda3.impl.TermReducerImpl;
import org.xmloperator.lambda3.impl.type.TermTypeFactory;
import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Function1;
import org.xmloperator.lambda3.model.Function2;
import org.xmloperator.lambda3.model.Function3;
import org.xmloperator.lambda3.model.Term;

/**
 * Usual constants.
 */
public abstract class Constants {

	// ////////////////////////////
	// Constants with one argument
	// ////////////////////////////

	private static final Function1 functionJ = new Function1() {
		public String getName() {
			return "J";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg) {
			return arg;
		}
	};
	private static final Function1 functionZ = new Function1() {
		public String getName() {
			return "Z";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg) {
			return constantJ;
		}
	};
	private static final Function1 functionE2 = new Function1() {
		public String getName() {
			return "E2";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg) {
			return newConstantT();
		}
	};
	private static final Function1 functionE3 = new Function1() {
		public String getName() {
			return "E3";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg) {
			return constantZ;
		}
	};
	private static final Function1 functionP = new Function1() {
		public String getName() {
			return "P";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg) {
			return arg.applyTo(constantP.applyTo(arg));
		}
	};
	private static final Function1 functionNorm = new Function1() {
		public String getName() {
			return "M";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg) {
			return new TermReducerImpl().normalForm(arg);
		}
	};

	/**
	 * Constant J such as JM -> M.
	 */
	public static final Constant constantJ = TermFactory.newConstant(functionJ,
			TermTypeFactory.identityFormula);

	/**
	 * Constant P such as PG -> G(PG).
	 */
	public static final Constant constantP = TermFactory.newConstant(functionP,
			TermTypeFactory.fixPointFormula);

	/**
	 * Constant Z such as ZM -> J.
	 */
	public static final Constant constantZ = TermFactory.newConstant(functionZ,
			TermTypeFactory.booleanFormula);

	/**
	 * Constant E2 such as (E2)M -> T.
	 */
	public static final Constant constantE2 = TermFactory.newConstant(
			functionE2, TermTypeFactory.trileanFormula);

	/**
	 * Constant E3 such as (E3)M -> Z.
	 */
	public static final Constant constantE3 = TermFactory.newConstant(
			functionE3, TermTypeFactory.trileanFormula);

	/**
	 * Constant Norm that normalizes its argument.
	 */
	public static final Constant constantNorm = TermFactory.newConstant(
			functionNorm, TermTypeFactory.identityFormula);

	// /////////////////////////////
	// Constants with two arguments
	// /////////////////////////////

	private static final Function2 functionT = new Function2() {
		public String getName() {
			return "T";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg1, Term arg2) {
			return arg1;
		}
	};

	/**
	 * Returns a new instance of the Constant T such as TMN -> M.
	 * 
	 * @return a new instance of the Constant T such as TMN -> M.
	 */
	public static final Constant newConstantT() {
		return TermFactory.newConstant(functionT,
				TermTypeFactory.booleanFormula);
	}

	/**
	 * Returns a new instance of a boolean Constant.
	 * 
	 * @param b
	 *            a boolean.
	 * @return a new instance of a boolean Constant.
	 */
	public static final Constant newConstantBoolean(boolean b) {
		return b ? newConstantT() : constantZ;
	}

	// ///////////////////////////////
	// Constants with three arguments
	// ///////////////////////////////

	private static final Function3 functionE1 = new Function3() {
		public String getName() {
			return "E1";
		}

		public boolean acceptArgument(int index, Term arg) {
			return true;
		}

		public Term function(Term arg1, Term arg2, Term arg3) {
			return arg1;
		}
	};

	/**
	 * Returns a new instance of the Constant E1 such as (E1)ABC -> A.
	 * 
	 * @return a new instance of the Constant E1 such as (E1)ABC -> A.
	 */
	public static final Constant newConstantE1() {
		return TermFactory.newConstant(functionE1,
				TermTypeFactory.trileanFormula);
	}

	/**
	 * Returns a new instance of a <i>trilean</i> Constant.
	 * 
	 * @param b
	 *            a <i>trilean</i> 1|2|3.
	 * @return a new instance of a <i>trilean</i> Constant.
	 */
	public static final Constant newConstantTrilean(int n) {
		if (n == 1)
			return newConstantE1();
		if (n == 2)
			return constantE2;
		if (n == 3)
			return constantE3;
		throw new IllegalArgumentException();
	}
}
