/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.coding;

import org.xmloperator.lambda3.exception.IllegalFunctionArgumentException;
import org.xmloperator.lambda3.impl.ConstantFloat;
import org.xmloperator.lambda3.impl.TermFactory;
import org.xmloperator.lambda3.impl.type.TermTypeFactory;
import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Function;
import org.xmloperator.lambda3.model.Function1;
import org.xmloperator.lambda3.model.Function2;
import org.xmloperator.lambda3.model.NumberCoding;
import org.xmloperator.lambda3.model.Term;

/**
 * Implementation of NumberCoding using float constants.
 */
public final class FloatNumberCoding implements NumberCoding {
	public static final NumberCoding NUMBER_CODING = new FloatNumberCoding();

	private static final double argValue(Function function, Term arg) {
		if (!arg.isConstant())
			throw new IllegalFunctionArgumentException(function, arg);
		Constant constant = (Constant) arg;
		if (!constant.isFloat())
			throw new IllegalFunctionArgumentException(function, arg);
		return ((ConstantFloat) constant).getValue();
	}

	private static final Function1 functionIsNegative = new Function1() {
		public String getName() {
			return "isNegative";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return Constants.newConstantBoolean(argValue(this, arg) < 0);
		}
	};
	private static final Function1 functionSqrt = new Function1() {
		public String getName() {
			return "sqrt";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantFloat(Math.sqrt(argValue(this, arg)));
		}
	};
	private static final Function1 functionLn = new Function1() {
		public String getName() {
			return "ln";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantFloat(Math.log(argValue(this, arg)));
		}
	};
	private static final Function1 functionExp = new Function1() {
		public String getName() {
			return "exp";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantFloat(Math.exp(argValue(this, arg)));
		}
	};
	private static final Function1 functionCos = new Function1() {
		public String getName() {
			return "cos";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantFloat(Math.cos(argValue(this, arg)));
		}
	};
	private static final Function1 functionSin = new Function1() {
		public String getName() {
			return "sin";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantFloat(Math.sin(argValue(this, arg)));
		}
	};
	private static final Function1 functionAtan = new Function1() {
		public String getName() {
			return "atan";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantFloat(Math.atan(argValue(this, arg)));
		}
	};
	private static final Function1 functionFromInteger = new Function1() {
		public String getName() {
			return "fromInteger";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantFloat(IntNumberCoding.argValue(this,
					arg));
		}
	};
	private static final Function1 functionToInteger = new Function1() {
		public String getName() {
			return "toInteger";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantInt((int) argValue(this, arg));
		}
	};

	private static final Function2 functionAdd = new Function2() {
		public String getName() {
			return "add";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantFloat(argValue(this, arg1)
					+ argValue(this, arg2));
		}
	};
	private static final Function2 functionSub = new Function2() {
		public String getName() {
			return "sub";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantFloat(argValue(this, arg1)
					- argValue(this, arg2));
		}
	};
	private static final Function2 functionMult = new Function2() {
		public String getName() {
			return "mult";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantFloat(argValue(this, arg1)
					* argValue(this, arg2));
		}
	};
	private static final Function2 functionDiv = new Function2() {
		public String getName() {
			return "div";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isFloat();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantFloat(argValue(this, arg1)
					/ argValue(this, arg2));
		}
	};

	public static final Constant constantIsNegative = TermFactory.newConstant(
			functionIsNegative, TermTypeFactory.floatToBooleanFormula);
	public static final Constant constantSqrt = TermFactory.newConstant(
			functionSqrt, TermTypeFactory.floatToFloatTermType);
	public static final Constant constantLn = TermFactory.newConstant(
			functionLn, TermTypeFactory.floatToFloatTermType);
	public static final Constant constantExp = TermFactory.newConstant(
			functionExp, TermTypeFactory.floatToFloatTermType);
	public static final Constant constantCos = TermFactory.newConstant(
			functionCos, TermTypeFactory.floatToFloatTermType);
	public static final Constant constantSin = TermFactory.newConstant(
			functionSin, TermTypeFactory.floatToFloatTermType);
	public static final Constant constantAtan = TermFactory.newConstant(
			functionAtan, TermTypeFactory.floatToFloatTermType);
	public static final Constant constantFromInteger = TermFactory.newConstant(
			functionFromInteger, TermTypeFactory.integerToFloatTermType);
	public static final Constant constantToInteger = TermFactory.newConstant(
			functionToInteger, TermTypeFactory.floatToIntegerTermType);

	public static final Constant newConstantAdd() {
		return TermFactory.newConstant(functionAdd,
				TermTypeFactory.floatToFloatToFloatTermType,
				TermTypeFactory.floatToFloatTermType);
	}

	public static final Constant newConstantSub() {
		return TermFactory.newConstant(functionSub,
				TermTypeFactory.floatToFloatToFloatTermType,
				TermTypeFactory.floatToFloatTermType);
	}

	public static final Constant newConstantMult() {
		return TermFactory.newConstant(functionMult,
				TermTypeFactory.floatToFloatToFloatTermType,
				TermTypeFactory.floatToFloatTermType);
	}

	public static final Constant newConstantDiv() {
		return TermFactory.newConstant(functionDiv,
				TermTypeFactory.floatToFloatToFloatTermType,
				TermTypeFactory.floatToFloatTermType);
	}

	// ////////////////////////////
	// NumberCoding implementation
	// ////////////////////////////

	public char getIntroductionChar() {
		return NumberCoding.FLOAT;
	}

	public Term codeNumber(boolean isSigned, int number) {
		return TermFactory.newConstantFloat(number);
	}

	public Term codeFunction(String functionName) {
		if (functionName.equals("isNegative"))
			return constantIsNegative;
		if (functionName.equals("sqrt"))
			return constantSqrt;
		if (functionName.equals("ln"))
			return constantLn;
		if (functionName.equals("exp"))
			return constantExp;
		if (functionName.equals("cos"))
			return constantCos;
		if (functionName.equals("sin"))
			return constantSin;
		if (functionName.equals("atan"))
			return constantAtan;
		if (functionName.equals("fromInteger"))
			return constantFromInteger;
		if (functionName.equals("toInteger"))
			return constantToInteger;
		if (functionName.equals(ADD))
			return newConstantAdd();
		if (functionName.equals(SUB))
			return newConstantSub();
		if (functionName.equals(MULT))
			return newConstantMult();
		if (functionName.equals(DIV))
			return newConstantDiv();
		return null;
	}
}
