/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.coding;

import org.xmloperator.lambda3.exception.IllegalFunctionArgumentException;
import org.xmloperator.lambda3.impl.ConstantInt;
import org.xmloperator.lambda3.impl.TermFactory;
import org.xmloperator.lambda3.impl.type.TermTypeFactory;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Function;
import org.xmloperator.lambda3.model.Function1;
import org.xmloperator.lambda3.model.Function2;
import org.xmloperator.lambda3.model.NumberCoding;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Variable;

/**
 * Implementation of NumberCoding using int constants.
 */
public final class IntNumberCoding implements NumberCoding {
	public static final NumberCoding NUMBER_CODING = new IntNumberCoding();

	static final int argValue(Function function, Term arg) {
		if (!arg.isConstant())
			throw new IllegalFunctionArgumentException(function, arg);
		Constant constant = (Constant) arg;
		if (!constant.isInt())
			throw new IllegalFunctionArgumentException(function, arg);
		return ((ConstantInt) constant).getValue();
	}

	private static final Function1 functionIsZero = new Function1() {
		public String getName() {
			return "_isZero";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			return Constants.newConstantBoolean(argValue(this, arg) == 0);
		}
	};
	private static final Function1 functionIsNegative = new Function1() {
		public String getName() {
			return "_isNegative";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			return Constants.newConstantBoolean(argValue(this, arg) < 0);
		}
	};
	private static final Function1 functionSucc = new Function1() {
		public String getName() {
			return "_succ";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantInt(argValue(this, arg) + 1);
		}
	};
	private static final Function1 functionPred = new Function1() {
		public String getName() {
			return "_pred";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			return TermFactory.newConstantInt(argValue(this, arg) - 1);
		}
	};
	private static final Function1 functionFromChurch = new Function1() {
		public String getName() {
			return "_fromChurch";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isAbstraction()
					&& ((Abstraction) arg).getChurchNumeral() >= 0;
		}

		public Term function(Term arg) {
			if (!arg.isAbstraction())
				throw new IllegalFunctionArgumentException(this, arg);
			int n = ((Abstraction) arg).getChurchNumeral();
			if (n < 0)
				throw new IllegalFunctionArgumentException(this, arg);
			return TermFactory.newConstantInt(n);
		}
	};
	private static final Function1 functionToChurch = new Function1() {
		public String getName() {
			return "_toChurch";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			int n = argValue(this, arg);
			if (n < 0)
				throw new IllegalFunctionArgumentException(this, arg);
			return ChurchNumberCoding.newNumber(n);
		}
	};
	private static final Function1 functionFromBinary = new Function1() {
		public String getName() {
			return "_fromBinary";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isAbstraction()
					&& BinaryNumberCoding.getValue((Abstraction) arg) >= 0;
		}

		public Term function(Term arg) {
			if (!arg.isAbstraction())
				throw new IllegalFunctionArgumentException(this, arg);
			int n = BinaryNumberCoding.getValue((Abstraction) arg);
			if (n < 0)
				throw new IllegalFunctionArgumentException(this, arg);
			return TermFactory.newConstantInt(n);
		}
	};
	private static final Function1 functionToBinary = new Function1() {
		public String getName() {
			return "_toBinary";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			int n = argValue(this, arg);
			if (n < 0)
				throw new IllegalFunctionArgumentException(this, arg);
			return BinaryNumberCoding.newNumber(n);
		}
	};
	private static final Function1 functionSyracuse = new Function1() {
		public String getName() {
			return "_syracuse";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg) {
			int n = argValue(this, arg);
			if (n % 2 == 1)
				n = 3 * n + 1;
			return TermFactory.newConstantInt(n / 2);
		}
	};

	private static final Function2 functionAdd = new Function2() {
		public String getName() {
			return "_add";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantInt(argValue(this, arg1)
					+ argValue(this, arg2));
		}
	};
	private static final Function2 functionSub = new Function2() {
		public String getName() {
			return "_sub";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantInt(argValue(this, arg1)
					- argValue(this, arg2));
		}
	};
	private static final Function2 functionMult = new Function2() {
		public String getName() {
			return "_mult";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantInt(argValue(this, arg1)
					* argValue(this, arg2));
		}
	};
	private static final Function2 functionDiv = new Function2() {
		public String getName() {
			return "_div";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantInt(argValue(this, arg1)
					/ argValue(this, arg2));
		}
	};
	private static final Function2 functionMod = new Function2() {
		public String getName() {
			return "_mod";
		}

		public boolean acceptArgument(int index, Term arg) {
			return arg.isConstant() && ((Constant) arg).isInt();
		}

		public Term function(Term arg1, Term arg2) {
			return TermFactory.newConstantInt(argValue(this, arg1)
					% argValue(this, arg2));
		}
	};

	public static final Constant constantIsZero = TermFactory.newConstant(
			functionIsZero, TermTypeFactory.integerToBooleanFormula);
	public static final Constant constantIsNegative = TermFactory.newConstant(
			functionIsNegative, TermTypeFactory.integerToBooleanFormula);
	public static final Constant constantSucc = TermFactory.newConstant(
			functionSucc, TermTypeFactory.integerToIntegerTermType);
	public static final Constant constantPred = TermFactory.newConstant(
			functionPred, TermTypeFactory.integerToIntegerTermType);
	public static final Constant constantFromChurch = TermFactory.newConstant(
			functionFromChurch, TermTypeFactory.churchIntToIntegerFormula);
	public static final Constant constantToChurch = TermFactory.newConstant(
			functionToChurch, TermTypeFactory.integerToChurchIntFormula);
	public static final Constant constantFromBinary = TermFactory.newConstant(
			functionFromBinary, TermTypeFactory.integerToIntegerTermType);
	public static final Constant constantToBinary = TermFactory.newConstant(
			functionToBinary, TermTypeFactory.integerToIntegerTermType);
	public static final Constant constantSyracuse = TermFactory.newConstant(
			functionSyracuse, TermTypeFactory.integerToIntegerTermType);

	public static final Constant newConstantAdd() {
		return TermFactory.newConstant(functionAdd,
				TermTypeFactory.integerToIntegerToIntegerTermType,
				TermTypeFactory.integerToIntegerTermType);
	}

	public static final Constant newConstantSub() {
		return TermFactory.newConstant(functionSub,
				TermTypeFactory.integerToIntegerToIntegerTermType,
				TermTypeFactory.integerToIntegerTermType);
	}

	public static final Constant newConstantMult() {
		return TermFactory.newConstant(functionMult,
				TermTypeFactory.integerToIntegerToIntegerTermType,
				TermTypeFactory.integerToIntegerTermType);
	}

	public static final Constant newConstantDiv() {
		return TermFactory.newConstant(functionDiv,
				TermTypeFactory.integerToIntegerToIntegerTermType,
				TermTypeFactory.integerToIntegerTermType);
	}

	public static final Constant newConstantMod() {
		return TermFactory.newConstant(functionMod,
				TermTypeFactory.integerToIntegerToIntegerTermType,
				TermTypeFactory.integerToIntegerTermType);
	}

	/**
	 * Returns a new instance of the operator that gives the step count of the
	 * Syracuse sequence starting at a given number.
	 */
	public static final Term newSyracuseStepCount() {
		/**
		 * n -> t(n) = stepCount
		 */
		Variable t = TermFactory.newVariable();
		Variable n = TermFactory.newVariable();
		/**
		 * g(t, n) = (if (isZero(pred(n))) then 0 else succ(t(syracuse(n))))
		 */
		Term g = constantIsZero.applyTo(constantPred.applyTo(n)).applyTo(
				TermFactory.newConstantInt(0)).applyTo(
				constantSucc.applyTo(t.applyTo(constantSyracuse.applyTo(n))))
				.abstractOn(n).abstractOn(t).setCombinatorName(
						"_syracuseStepCount");
		/**
		 * return PG, PG = t = G(t)
		 */
		return Constants.constantP.applyTo(g);
	}

	// ////////////////////////////
	// NumberCoding implementation
	// ////////////////////////////

	public char getIntroductionChar() {
		return NumberCoding.INT;
	}

	public Term codeNumber(boolean isSigned, int number) {
		return TermFactory.newConstantInt(number);
	}

	public Term codeFunction(String functionName) {
		if (functionName.equals(ISZERO))
			return constantIsZero;
		if (functionName.equals("isNegative"))
			return constantIsNegative;
		if (functionName.equals(SUCC))
			return constantSucc;
		if (functionName.equals(PRED))
			return constantPred;
		if (functionName.equals(ADD))
			return newConstantAdd();
		if (functionName.equals(SUB))
			return newConstantSub();
		if (functionName.equals(MULT))
			return newConstantMult();
		if (functionName.equals(DIV))
			return newConstantDiv();
		if (functionName.equals(MOD))
			return newConstantMod();
		if (functionName.equals("fromChurch"))
			return constantFromChurch;
		if (functionName.equals("toChurch"))
			return constantToChurch;
		if (functionName.equals("fromBinary"))
			return constantFromBinary;
		if (functionName.equals("toBinary"))
			return constantToBinary;
		if (functionName.equals("syracuse"))
			return constantSyracuse;
		if (functionName.equals("syracuseStepCount"))
			return newSyracuseStepCount();
		return null;
	}
}
