/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.type;

import org.xmloperator.lambda3.model.type.ConstantTermType;
import org.xmloperator.lambda3.model.type.TermType;
import org.xmloperator.lambda3.model.type.TermTypeFormula;

/**
 * TermType factory.
 */
public abstract class TermTypeFactory {

	public static final TermType integerTermType = new ConstantTermTypeImpl(
			ConstantTermType.CHAR_INTEGER_TYPE_NAME);
	public static final TermType integerToIntegerTermType = newFunctionTermType(
			integerTermType, integerTermType);
	public static final TermType integerToIntegerToIntegerTermType = newFunctionTermType(
			integerTermType, integerToIntegerTermType);

	public static final TermType floatTermType = new ConstantTermTypeImpl(
			ConstantTermType.CHAR_FLOAT_TYPE_NAME);
	public static final TermType floatToFloatTermType = newFunctionTermType(
			floatTermType, floatTermType);
	public static final TermType floatToFloatToFloatTermType = newFunctionTermType(
			floatTermType, floatToFloatTermType);
	public static final TermType integerToFloatTermType = newFunctionTermType(
			integerTermType, floatTermType);
	public static final TermType floatToIntegerTermType = newFunctionTermType(
			floatTermType, integerTermType);

	public static final TermTypeFormula identityFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newIdentityTermType(variableTermType);
		}
	};
	public static final TermTypeFormula fixPointFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newFixPointTermType(variableTermType);
		}
	};
	public static final TermTypeFormula booleanFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newBooleanTermType(variableTermType);
		}
	};
	public static final TermTypeFormula trileanFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newTrileanTermType(variableTermType);
		}
	};
	public static final TermTypeFormula churchIntFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newChurchIntTermType(variableTermType);
		}
	};
	public static final TermTypeFormula churchIntToIntegerFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newFunctionTermType(newChurchIntTermType(variableTermType),
					integerTermType);
		}
	};
	public static final TermTypeFormula integerToChurchIntFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newFunctionTermType(integerTermType,
					newChurchIntTermType(variableTermType));
		}
	};
	public static final TermTypeFormula integerToBooleanFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newFunctionTermType(integerTermType,
					newBooleanTermType(variableTermType));
		}
	};
	public static final TermTypeFormula floatToBooleanFormula = new TermTypeFormula() {
		public TermType termTypeFormula(TermType variableTermType) {
			return newFunctionTermType(floatTermType,
					newBooleanTermType(variableTermType));
		}
	};

	/**
	 * Returns a new variable TermType.
	 * 
	 * @return a new variable TermType.
	 */
	public static final TermType newVariableTermType() {
		return new VariableTermTypeImpl();
	}

	/**
	 * Returns a new function TermType.
	 * 
	 * @param input
	 *            a TermType as input.
	 * @param output
	 *            a TermType as output.
	 * @return a new function TermType.
	 */
	public static final TermType newFunctionTermType(TermType input,
			TermType output) {
		return new FunctionTermTypeImpl((TermTypeImpl) input,
				(TermTypeImpl) output);
	}

	/**
	 * Returns a new abstraction of a TermType.
	 * 
	 * @param argTermType
	 *            a TermType to abstract.
	 * @param abstractionCount
	 *            a number of abstraction.
	 * @return the new TermType.
	 */
	public static final TermType newAbstractionTermType(TermType argTermType,
			int abstractionCount) {
		TermType termType = argTermType;
		for (int k = 0; k < abstractionCount; k++)
			termType = newFunctionTermType(newVariableTermType(), termType);
		return termType;
	}

	/**
	 * Returns a new TermType, which represents T->T.
	 * 
	 * @param t
	 *            a base TermType.
	 * @return a new TermType, which represents T->T.
	 */
	public static final TermType newIdentityTermType(TermType t) {
		return newFunctionTermType(t, t);
	}

	/**
	 * Returns a new TermType, which represents (T->T)->T.
	 * 
	 * @param t
	 *            a base TermType.
	 * @return a new TermType, which represents (T->T)->T.
	 */
	public static final TermType newFixPointTermType(TermType t) {
		return newFunctionTermType(newIdentityTermType(t), t);
	}

	/**
	 * Returns a new TermType, which represents T->T->T.
	 * 
	 * @param t
	 *            a base TermType.
	 * @return a new TermType, which represents T->T->T.
	 */
	public static final TermType newBooleanTermType(TermType t) {
		return newFunctionTermType(t, newIdentityTermType(t));
	}

	/**
	 * Returns a new TermType, which represents T->T->T->T.
	 * 
	 * @param t
	 *            a base TermType.
	 * @return a new TermType, which represents T->T->T->T.
	 */
	public static final TermType newTrileanTermType(TermType t) {
		return newFunctionTermType(t, newBooleanTermType(t));
	}

	/**
	 * Returns a new TermType, which represents (T->T)->T->T.
	 * 
	 * @param t
	 *            a base TermType.
	 * @return a new TermType, which represents (T->T)->T->T.
	 */
	public static final TermType newChurchIntTermType(TermType t) {
		TermType i = newIdentityTermType(t);
		return newFunctionTermType(i, i);
	}

	/**
	 * Returns a new TermType, which represents
	 * ((T1->T1)->T1->T1)->(T2->T2)->T2->T2.
	 * 
	 * @param t1
	 *            a first base TermType.
	 * @param t1
	 *            a second base TermType.
	 * @return a new TermType, which represents
	 *         ((T1->T1)->T1->T1)->(T2->T2)->T2->T2.
	 */
	public static final TermType newChurchIntToChurchIntTermType(TermType t1,
			TermType t2) {
		return TermTypeFactory.newFunctionTermType(newChurchIntTermType(t1),
				newChurchIntTermType(t2));
	}
}
