/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.model;

import java.util.Set;

public interface TermWriter {

	/**
	 * Returns a text representation of a root Term using the indexed notation
	 * of De Bruijn. If the underlinedTerm argument is not null then returns an
	 * underscore of the underlinedTerm within the root Term.
	 * 
	 * <pre>
	 * Example : "\1".
	 * </pre>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param underlinedTerm
	 *            an optional sub-term to underline.
	 * @return a text representation of a root Term using the indexed notation
	 *         of De Bruijn. If the underlinedTerm argument is not null then
	 *         returns an underscore of the underlinedTerm within the root Term.
	 */
	public String indexedText(Term rootTerm, Term underlinedTerm);

	/**
	 * Returns a text representation of a root Term using the indexed notation
	 * of De Bruijn.
	 * 
	 * <pre>
	 * Example : "\1".
	 * </pre>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @return a text representation of a root Term using the indexed notation
	 *         of De Bruijn.
	 */
	public String indexedText(Term rootTerm);

	/**
	 * Returns a text representation of a root Term as usual lambda term. If the
	 * underlinedTerm argument is not null then returns an underscore of the
	 * underlinedTerm within the root Term.
	 * 
	 * <pre>
	 * Example : "\x1.x1".
	 * </pre>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param underlinedTerm
	 *            an optional sub-term to underline.
	 * @return a text representation of a root Term as usual lambda term. If the
	 *         underlinedTerm argument is not null then returns an underscore of
	 *         the underlinedTerm within the root Term.
	 */
	public String lambdaText(Term rootTerm, Term underlinedTerm);

	/**
	 * Returns a text representation of a root Term as usual lambda term.
	 * 
	 * <pre>
	 * Example : "\x1.x1".
	 * </pre>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @return a text representation of a root Term as usual lambda term.
	 */
	public String lambdaText(Term rootTerm);

	/**
	 * Returns a root Term as a DAG. if the underlinedTerm argument is not null
	 * then returns an underscore of the underlinedTerm within the root Term.
	 * 
	 * <pre>
	 * Example(Omega): "A1A1 A1=\x1.x1x1".
	 * </pre>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param underlinedTerm
	 *            an optional sub-term to underline.
	 * @return a text representation of a root Term as a DAG. if the
	 *         underlinedTerm argument is not null then returns an underscore of
	 *         the underlinedTerm within the root Term.
	 */
	public String dagText(Term rootTerm, Term underlinedTerm);

	/**
	 * Returns a root Term as a DAG.
	 * 
	 * <pre>
	 * Example(Omega): "A1A1 A1=\x1.x1x1".
	 * </pre>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @return a text representation of a root Term as a DAG.
	 */
	public String dagText(Term rootTerm);

	/**
	 * Returns an underscore of the given redex within a root Term.
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param redex
	 *            a redex within the root Term.
	 * @return an underscore of the given redex within the given root Term.
	 */
	public String dagRedexUnderscore(Term rootTerm, Application redex);

	/**
	 * Returns an underscore of the given Abstraction set within a root Term.
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param abstractionSet
	 *            an Abstraction set within the root Term.
	 * @return an underscore of the given Abstraction set within a root Term.
	 */
	public String dagAbstractionSetUnderscore(Term rootTerm,
			Set<Abstraction> abstractionSet);
}
